// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// body.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class bodies                                                                |
// class body                                                                  |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_body_h
#define included_body_h 1

#ifndef included_auxx_h
#include <tree/auxx.h>
#endif

#ifndef included_flag_h
#include <tree/flag.h>
#endif

#ifdef ALLOW_NEMO
#ifndef included_nmio_h
#include <tree/nmio.h>
#endif
#endif

////////////////////////////////////////////////////////////////////////////////
namespace fasttree {
  class body;                                      // forward declaration       
  class soul;                                      // forward declaration       
  class nemo_input;                                // forward declaration       
  class nemo_output;                               // forward declaration       
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::bodies                                                     
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // - allocates bodies in one block                                            
  // - provides iterators (class body below) for sequential access to all bodies
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  class bodies {
    //--------------------------------------------------------------------------
    // friendship                                                               
    //--------------------------------------------------------------------------
    friend class body;                             // for data access           
  private:
    bodies(const bodies&);                         // not implemented           
    bodies& operator=(const bodies&);              // not implemented           
    //--------------------------------------------------------------------------
    // public types                                                             
    //--------------------------------------------------------------------------
  public:
    typedef class body body_type;                  // type of body              
    //--------------------------------------------------------------------------
    // private types                                                            
    //--------------------------------------------------------------------------
  private:
    enum {
      NOT_LONGER  = 32,                            // use 6th bit of basic_flag 
      NOT_SHORTER = 64                             // use 7th bit of basic_flag 
    };
    //--------------------------------------------------------------------------
    struct iodata {                                // I/O data of a body        
      real         MASS;                           //   mass                    
      vect         POS;                            //   position                
      vect         VEL;                            //   velocity                
      real         EPS;                            //   softening length        
    };      
    //--------------------------------------------------------------------------
    struct basic_body : public basic_flag {        // elementary data of a body 
      iodata       IOD;                            //   I/O data                
      mutable soul*SOUL;                           //   soul for gravity etc    
      real &mass() { return IOD.MASS; }
      vect &pos () { return IOD.POS; }
      vect &vel () { return IOD.VEL; }
      real &eps () { return IOD.EPS; }
    };
    //--------------------------------------------------------------------------
    struct blockstep_support {                     // data used by blockstepping
      indx         LEVEL;                          //   time-step level         
      vect         OPOS;                           //   old position            
      vect         OACC;                           //   old acceleration        
    };
    //--------------------------------------------------------------------------
    struct sph_support {                           // data used by SPH etc      
      real         SIZE;                           //   SPH size of body        
    };
    //--------------------------------------------------------------------------
    // data members                                                             
    //--------------------------------------------------------------------------
                                                   // 1 data of all bodies      
    uint                N, NSPH;                   //    # bodies in total      
                                                   // 2 data of bodies          
    basic_body         *BB;                        //    basic properties       
    mutable
    blockstep_support  *BS;                        //    support for blocksteps 
    sph_support        *SS;                        //    support for SPH        
    //--------------------------------------------------------------------------
    // construction & destruction                                               
    //--------------------------------------------------------------------------
  public:
    bodies() : N(0), NSPH(0), BB(0), BS(0), SS(0) {}
    //--------------------------------------------------------------------------
    bodies(                                        // constructor               
	   const uint n,                           // I: # bodies               
	   const bool e    = false,                //[I: add stepping support]  
	   const uint nsph = 0u) :                 //[I: first N_sph: SPH]      

      N     ( n ),
      NSPH  ( min(n,nsph) ),
      BB    ( N?        new basic_body[N]        : 0 ),
      BS    ( (e && N)? new blockstep_support[N] : 0 ),
      SS    ( NSPH ?    new sph_support[NSPH]    : 0 )
    {
      if(NSPH)
	for(register basic_body*Bi=BB; Bi!=BB+NSPH; ++Bi)
	  Bi->add_flag(basic_flag::SPH);
    }
    //--------------------------------------------------------------------------
#ifdef ALLOW_NEMO
    bodies(                                        // constructor               
	   char* f,                                // I: from this NEMO snapshot
	   real* t= 0) :                           //[I: read time?]            
      N(0), NSPH(0), BB(0), BS(0), SS(0)
    {
      read_nemo_snapshot(f,t);
    }
#endif
    //--------------------------------------------------------------------------
    void reset(                                    // change N & NSPH           
	       const uint n,                       // I: # bodies               
	       const uint nsph = 0u)               //[I: first N_sph: SPH]      
    {
      if(nsph != NSPH) {
	NSPH = min(nsph,n);
	if(SS) delete[] SS;
	SS = NSPH ? new sph_support[NSPH] : 0;
      }
      if(n != N) {
	N = n;
	if(BB) delete[] BB;
	BB = N ? new basic_body[N] : 0;
	for(register basic_body*Bi=BB; Bi!=BB+NSPH; ++Bi)
	  Bi->add_flag(basic_flag::SPH);
	if(BS) {
	  delete[] BS;
	  BS = N ? new blockstep_support[N] : 0;
	}
      }
    }
    //--------------------------------------------------------------------------
    ~bodies() {
      if(BB) delete[] BB;
      if(BS) delete[] BS;
      if(SS) delete[] SS;
    }
    //--------------------------------------------------------------------------
    // I/O                                                                      
    //--------------------------------------------------------------------------
#ifdef ALLOW_NEMO
    void read_nemo_snapshot  (char*,                   // I: file for input     
			      real*     =0);           //[I: read time]         
    void write_nemo_snapshot (char*,                   // I: file for output    
			      const bool=false,        //[I: write potential ?] 
			      const bool=false,        //[I: write density ?]   
			      const real=0) const;     //[I: write time]        
    void read_nemo_particles (nemo_input*,             // I: nemo input:  nmio.h
			      real*     =0);           //[I: read time]         
    void write_nemo_particles(nemo_output*,            // I: nemo output: nmio.h
			      const bool=false,        //[I: write potential ?] 
			      const bool=false,        //[I: write density ?]   
			      const real=0) const;     //[I: write time]        
#endif
    void write_ascii         (ostream&,                // I: output stream      
			      const bool=false,        //[I: output eps_i ?]    
			      const bool=false,        //[I: output pot_i ?]    
			      const bool=false) const; //[I: output rho_i ?]    
    void read_ascii          (istream&,                // I: input stream       
			      const bool=false);       //[I: input eps_i ?]     
    void write_binary        (ostream&) const;         // I: output stream      
    void read_binary         (istream&);               // I: input stream       
    //--------------------------------------------------------------------------
    // direct data access                                                       
    //                                                                          
    // to be used instead of the arrays themselves                              
    // (this is to simplify possible changes in memory layout: they only affect 
    // these few inline functions)                                              
    //--------------------------------------------------------------------------
    basic_flag &flag  (const uint&i) const { return BB[i]; }
    real       &mass  (const uint&i) const { return BB[i].IOD.MASS; }
    vect       &pos   (const uint&i) const { return BB[i].IOD.POS; }
    vect       &vel   (const uint&i) const { return BB[i].IOD.VEL; }
    real       &eps   (const uint&i) const { return BB[i].IOD.EPS; }
    soul*      &mysoul(const uint&i) const { return BB[i].SOUL; }
    vect       &opos  (const uint&i) const { return BS[i].OPOS; }
    vect       &oacc  (const uint&i) const { return BS[i].OACC; }
    indx       &level (const uint&i) const { return BS[i].LEVEL; }
    real       &size  (const uint&i) const { return SS[i].SIZE; }
    vect       &acc   (const uint& ) const;
    real       &pot   (const uint& ) const;
    real       &rho   (const uint& ) const;
    const vect &c_acc (const uint& ) const;
    const real &c_pot (const uint& ) const;
    const real &c_rho (const uint& ) const;
    //--------------------------------------------------------------------------
    // simple manipulations                                                     
    //--------------------------------------------------------------------------
    void set_soul(soul* const &S, const uint&i) const { BB[i].SOUL = S; }
    //--------------------------------------------------------------------------
    // allocation & de-allocation                                               
    //--------------------------------------------------------------------------
    void enable_time_steps() const {
      if(BS==0 && N!=0) BS  = new blockstep_support[N];
    }
    //--------------------------------------------------------------------------
    void disable_time_steps() const {
      if(BS) { delete[] BS; BS=0; }
    }
    //--------------------------------------------------------------------------
    // constant methods                                                         
    //--------------------------------------------------------------------------
    const uint& N_body            () const { return N; }
    const uint& N_SPH             () const { return NSPH; }
    bool        support_timesteps () const { return BS; }
    bool        support_SPH       () const { return NSPH; }
    //--------------------------------------------------------------------------
    // access to bodies (inlined after definition of class fasttree::body)      
    //--------------------------------------------------------------------------
    body body_no     (const uint&) const;
    body begin_bodies() const;
    body end_bodies  () const;
    body back_bodies () const;
    body begin_SPH   () const;
    body end_SPH     () const;
    body back_SPH    () const;
  };
}
////////////////////////////////////////////////////////////////////////////////
#ifndef included_soul_h
#include <tree/soul.h>
#endif
////////////////////////////////////////////////////////////////////////////////
namespace fasttree {
  //----------------------------------------------------------------------------
  inline vect&bodies::acc(const uint&i) const { return BB[i].SOUL->acc(); }
  inline real&bodies::pot(const uint&i) const { return BB[i].SOUL->pot(); }
  inline real&bodies::rho(const uint&i) const { return BB[i].SOUL->rho(); }
  //----------------------------------------------------------------------------
  inline const vect&bodies::c_acc(const uint&i) const {
    return fasttree::acc(BB[i].SOUL); }
  inline const real&bodies::c_pot(const uint&i) const {
    return fasttree::pot(BB[i].SOUL); }
  inline const real&bodies::c_rho(const uint&i) const {
    return fasttree::rho(BB[i].SOUL); }
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::body                                                       
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // - serves like a data class body, in that it gives direct access (const and 
  //   non-const) to the data members of a body                                 
  // - also works like a forward iterator on bodys, using increment:            
  //   for(body b=bodies::begin(); b!=bodies::end(); ++b) { ... }               
  // - cannot be publicly constructed from nothing                              
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  class body {
    //--------------------------------------------------------------------------
    // friendship to enable construction of a body                              
    //--------------------------------------------------------------------------
  private:
    friend body bodies::body_no     (const uint&) const;
    friend body bodies::begin_bodies() const;
    friend body bodies::end_bodies()   const;
    friend body bodies::back_bodies()  const;
    friend body bodies::end_SPH()      const;
    friend body bodies::back_SPH()     const;
    //--------------------------------------------------------------------------
    // data members                                                             
    //--------------------------------------------------------------------------
    const bodies *M;                               // my bodies                 
    uint          I;                               // my index                  
    //--------------------------------------------------------------------------
    // construction                                                             
    //--------------------------------------------------------------------------
  protected:
    body() {}
    body(const bodies*m, uint i) : M(m), I(i) {}
  public:
    body(const body&B) : M(B.M), I(B.I) {}
    body& operator= (const body&B) { M=B.M; I=B.I; return *this; }
    //--------------------------------------------------------------------------
    // iteration: only forward                                                  
    //--------------------------------------------------------------------------
    body& operator++()    { ++I; return *this; }   // prefix ++                 
    body  operator++(int) { return body(M,I++); }  // postfix ++                
    body& operator+=(const uint&i) { I+=i; return *this; }
    body  operator+ (const uint&i) const { return body(M,I+i); }
    bool  operator==(const body&B) const { return I == B.I; }
    bool  operator!=(const body&B) const { return I != B.I; }
    bool  operator< (const body&B) const { return I <  B.I; }
    bool  operator<=(const body&B) const { return I <= B.I; }
    bool  operator> (const body&B) const { return I >  B.I; }
    bool  operator>=(const body&B) const { return I >= B.I; }
    //--------------------------------------------------------------------------
    // de-referencing                                                           
    //                                                                          
    // for old code, using pointers to body (defined to be actually hold the    
    // data), to be made more easily compatible.                                
    // Don't use in new code.                                                   
    //--------------------------------------------------------------------------
    body& operator* () { return *this; }
    body* operator->() { return  this; }
    //--------------------------------------------------------------------------
    // flag manipulations                                                       
    //--------------------------------------------------------------------------
    bool flag_is_set   (const flag_type&F) const { return M->flag(I).is_set(F); }
    operator const basic_flag& ()          const { return M->flag(I); }
    void flag_as_sink      () { M->flag(I).add_flag(basic_flag::SINK); }
    void unflag_sink       () { M->flag(I).un_flag (basic_flag::SINK); }
    void flag_as_sticky    () { M->flag(I).add_flag(basic_flag::STICKY); }
    void unflag_sticky     () { M->flag(I).un_flag (basic_flag::STICKY); }
    void flag_as_sph       () { M->flag(I).add_flag(basic_flag::SPH); }
    void unflag_sph        () { M->flag(I).un_flag (basic_flag::SPH); }
    void allow_tree_usage  () { M->flag(I).un_flag (basic_flag::NOT_IN_TREE); }
    void forbid_tree_usage () { M->flag(I).add_flag(basic_flag::NOT_IN_TREE); }
    void allow_longer      () { M->flag(I).un_flag (bodies::NOT_LONGER); }
    void forbid_longer     () { M->flag(I).add_flag(bodies::NOT_LONGER); }
    void allow_shorter     () { M->flag(I).un_flag (bodies::NOT_SHORTER); }
    void forbid_shorter    () { M->flag(I).add_flag(bodies::NOT_SHORTER); }
    //--------------------------------------------------------------------------
    // non-const data access via members                                        
    //--------------------------------------------------------------------------
    real&mass ()            { return M->mass(I); }
    vect&pos  ()            { return M->pos(I); }
    real&pos  (const int i) { return M->pos(I)[i]; }
    vect&opos ()            { return M->opos(I); }
    real&opos (const int i) { return M->opos(I)[i]; }
    vect&vel  ()            { return M->vel(I); }
    real&vel  (const int i) { return M->vel(I)[i]; }
    vect&acc  ()            { return M->acc(I); }
    real&acc  (const int i) { return M->acc(I)[i]; }
    vect&oacc ()            { return M->oacc(I); }
    real&oacc (const int i) { return M->oacc(I)[i]; }
    real&eps  ()            { return M->eps(I); }
    real&pot  ()            { return M->pot(I); }
    real&rho  ()            { return M->rho(I); }
    indx&level()            { return M->level(I); }
    real&size ()            { return M->size(I); }
    //--------------------------------------------------------------------------
    // const data access via members                                            
    //--------------------------------------------------------------------------
    basic_flag const&flag () const { return M->flag(I); } 
    real       const&mass () const { return M->mass(I); }
    vect       const&pos  () const { return M->pos(I); }
    vect       const&opos () const { return M->opos(I); }
    vect       const&vel  () const { return M->vel(I); }
    vect       const&acc  () const { return M->c_acc(I); }
    vect       const&oacc () const { return M->oacc(I); }
    real       const&eps  () const { return M->eps(I); }
    real       const&pot  () const { return M->c_pot(I); }
    real       const&rho  () const { return M->c_rho(I); }
    indx       const&level() const { return M->level(I); }
    real       const&size () const { return M->size(I); }
    //--------------------------------------------------------------------------
    // const boolean informations via members                                   
    //--------------------------------------------------------------------------
    bool is_source     () const { return mass() != zero; }
    bool may_go_longer () const { return !flag_is_set(bodies::NOT_LONGER);}
    bool may_go_shorter() const { return !flag_is_set(bodies::NOT_SHORTER); }
    //--------------------------------------------------------------------------
    // const data access via friends                                            
    //--------------------------------------------------------------------------
#define CBB const body&B
    friend basic_flag   const&flag   (CBB) { return B.flag(); } 
    friend uint         const&index  (CBB) { return B.I; }
    friend const bodies*const&mygroup(CBB) { return B.M; }
    friend real         const&mass   (CBB) { return B.mass(); }
    friend vect         const&pos    (CBB) { return B.pos(); }
    friend vect         const&opos   (CBB) { return B.opos(); }
    friend vect         const&vel    (CBB) { return B.vel(); }
    friend vect         const&acc    (CBB) { return B.acc(); }
    friend vect         const&oacc   (CBB) { return B.oacc(); }
    friend real         const&eps    (CBB) { return B.eps(); }
    friend real         const&pot    (CBB) { return B.pot(); }
    friend real         const&rho    (CBB) { return B.rho(); }
    friend indx         const&level  (CBB) { return B.level(); }
    friend real         const&size   (CBB) { return B.size(); }
    friend real               energy (CBB) { return B.pot()+0.5*norm(B.vel()); }
    friend amom               angmom (CBB) { return (B.pos()) ^ (B.vel()); }
    //--------------------------------------------------------------------------
    // const boolean informations via friends                                   
    //--------------------------------------------------------------------------
    friend bool is_source     (CBB){return B.is_source(); }
    friend bool may_go_longer (CBB){return B.may_go_longer(); }
    friend bool may_go_shorter(CBB){return B.may_go_shorter(); }
#undef CBB
    //--------------------------------------------------------------------------
    // output: give just the index (a body acts like a pointer)                 
    //--------------------------------------------------------------------------
    friend ostream& operator<<(ostream&O, const body&B) { return O<<(B.I); }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::bodies: some inline member functions                       
  //////////////////////////////////////////////////////////////////////////////
  inline body bodies::body_no     (const uint&i) const { return body(this,i); }
  inline body bodies::begin_bodies() const { return body(this,0); }
  inline body bodies::end_bodies  () const { return body(this,N); }
  inline body bodies::back_bodies () const { return body(this,N-1); }
  inline body bodies::begin_SPH   () const { return begin_bodies(); }
  inline body bodies::end_SPH     () const { return body(this,NSPH); }
  inline body bodies::back_SPH    () const { return body(this,NSPH-1); }
}
////////////////////////////////////////////////////////////////////////////////
// useful macros                                                                
////////////////////////////////////////////////////////////////////////////////
#ifndef  LoopBodies
#define LoopBodies(BODIES_TYPE,                   /* type of bodies         */	\
		   BODIES_PTER,                   /* pointer to bodies      */	\
		   NAME)                          /* name for body          */	\
  for(register BODIES_TYPE::body_type             /* type of body           */	\
      NAME  = BODIES_PTER->begin_bodies();        /* from first body        */	\
      NAME != BODIES_PTER->end_bodies();          /* until beyond last body */	\
    ++NAME)                                       /* get next body          */
#endif
//------------------------------------------------------------------------------
#ifndef  LoopSPHBodies
#define LoopSPHBodies(BODIES_TYPE,                /* type of bodies         */	\
		      BODIES_PTER,                /* pointer to bodies      */	\
		      NAME)                       /* name for body          */	\
  for(register BODIES_TYPE::body_type             /* type of body           */	\
      NAME  = BODIES_PTER->begin_SPH();           /* from first body        */	\
      NAME != BODIES_PTER->end_SPH();             /* until beyond last body */	\
    ++NAME)                                       /* get next body          */
#endif
////////////////////////////////////////////////////////////////////////////////
#endif // included_body_h
