// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//  version 3.4.0                                                              |
//                                                                             |
//  falcON.h                                                                   |
//                                                                             |
//=============================================================================+
//                                                                             |
//  FALCON = Force ALgorithm with Complexity O(N)                              |
//                                                                             |
//=============================================================================+
//                                                                             |
//  header file for C++ users                                                  |
//  (C and FORTRAN users, see files falcON_C.h and falcON.f, respectively)     |
//                                                                             |
//  Copyright Walter Dehnen, 2000/2001                                         |
//  e-mail:   dehnen@mpia.de                                                   |
//  address:  Max-Planck Institut fr Astronomie,                              |
//            Knigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//  C++ source code implementing the code described by Dehnen (2000,2001b).    |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//  See file readme.falcON for some guidelines on how to install, #include and |
//  link this file and the code declared in it. You have the choice between    |
//  various options determining some properties of the code.                   |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//  CONTENTS                                                                   |
//  ========                                                                   |
//                                                                             |
//  1  Initialisation and clearing up                                          |
//     1.1  Before using the code                                              |
//     1.2  After using the code                                               |
//  2  Meaning of the body flags                                               |
//  3  Force approximation                                                     |
//     3.1  Generating a tree structure                                        |
//     3.2  Approximating the forces                                           |
//     3.3  Estimating mass- and number-density                                |
//  4  Search for Neighbours and Collision Partners                            |
//     4.1  Neighbour search (SPH support)                                     |
//     4.2  Collision partner search (for sticky particles)                    |
//     4.3  SPH Neighbour counting                                             |
//  5  Other features                                                          |
//  6  Known bugs and problems                                                 |
//     6.1  Test-bodies are not possible                                       |
//     6.2  Bodies at identical positions                                      |
//  References                                                                 |
//  A  Non-public code and code used by the C and FORTRAN support              |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//  See the detailled comments below for how to use this code.                 |
//  Some abbreviations:                                                        |
//                                                                             |
//  R:       return value                                                      |
//  I:       function argument is input                                        |
//  O:       function argument is output                                       |
//  I/O:     function argument is input & output                               |
//  [I]:     function argument is optional                                     |
//                                                                             |
//  Note that in case of optional function arguments, one may only omit the    |
//  last one (and the second but last if the last is already omitted, etc).    |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//  This code uses types defined in file auxx.h (via typedefs), which are:     |
//                                                                             |
//  real     a single or double precision floating point number, depending on  |
//           the macro DOUBLEPRECISION                                         |
//  areal    a single or double precision floating point number, depending on  |
//           the macro DOUBLEINPUT                                             |
//  vect     a tripel (for 3D) or pair (for 2D) of reals                       |
//  indx     an unsigned short (for holding small unsigned integers)           |
//  uint     an unsigned integer                                               |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_falcON_h                  // ensure this file is seen once    |
#define included_falcON_h 1                // only by the compiler             |
                                           //                                  |
#ifndef included_iostream                  //                                  |
#include <iostream>                        // basic I/O                        |
#define included_iostream                  //                                  |
#endif                                     //                                  |
                                           //                                  |
#ifndef included_auxx_h                    //                                  |
#include <tree/auxx.h>                     // basic types etc.                 |
#endif                                     //                                  |
                                           //                                  |
#ifndef included_kern_h                    //                                  |
#include <tree/kern.h>                     // enums kern_type & soft_type      |
#endif                                     //                                  |
//-----------------------------------------------------------------------------+
namespace fasttree {                       // falcON is in namespace fasttree  |
  template<class A> class tree;            // forward declaration of some      |
  class bodies;                            // classes and types                |
  class soul;                              // that class falcON holds          |
  class cell;                              // (we avoid to #include them       |
  class opening;                           // in order to insulate them from   |
  class TreeStats;                         // the clients of class falcON)     |
  class TaylorCoeffs;                      // (we avoid to #include them       |
//-----------------------------------------------------------------------------+
  class falcON {                           // Force ALgorithm with             |
  public:                                  // Complexity O(N)                  |
//-----------------------------------------------------------------------------+
//                                                                             |
//                                                                             |
//  1 INITIALISATION AND RELATED                                               |
//  ============================                                               |
//                                                                             |
//  1.1 CONSTRUCTORS                                                           |
//  ----------------                                                           |
//  In order to use the code, one has first to construct an object of type     |
//  `falcON'. This is done in C++ by the statement                             |
//                                                                             |
//      fasttree::falcON my_falcon( argument_list );                           |
//                                                                             |
//  which invokes a constructor for class falcON. There are, in fact two       |
//  constructors, one for the use with bodies as defined in file body.h:       |
//                                                                             |
    falcON   (const bodies*,               // I:   bodies                      |
	      const real,                  // I:   eps = softening length      |
	      const real      = 0.4,       // [I]: theta = opening angle       |
	      const kern_type = p1,        // [I]: type of softening kernel    |
	      const unsigned  = 50);       // [I]: max allowd tree depth       |
//                                                                             |
//  and one for the use with plain arrays:                                     |
//                                                                             |
    falcON   (const int*,                  // I:   array with flags            |
	      const areal*,                // I:   array with masses           |
	      const areal* [NDIM],         // I:   array with positions        |
	      const areal*,                // I:   array with soften. lengths  |
	            areal* [NDIM],         // I:   array for accelerations     |
	            areal*,                // I:   array for potentials        |
	            areal*,                // I:   array for densities         |
	      const unsigned,              // I:   N   = size of arrays        |
	      const real,                  // I:   eps = softening length      |
	      const real      = 0.4,       // [I]: theta = opening angle       |
	      const kern_type = p1,        // [I]: type softening kernel       |
	      const unsigned  = 50);       // [I]: max tree depth              |
//                                                                             |
//  This second version is designed for the use with C and FORTRAN, see files  |
//  falcON_C.h and falcON.f. Once a `falcON' is constructed, one can no longer |
//  switch between using bodies or arrays. Similarly, one cannot change the    |
//  arrays used, but, of course, one can change the data stored in these       |
//  arrays. Using the flags, one can in this way obtain almost any desired     |
//  behaviour, see below.                                                      |
//                                                                             |
//  The first argument(s) specify the sink and source properties of the        |
//  bodies. Each body has the sink properties: position (x,y,z), mass,         |
//  softening length (for the case of individual softening lengths), and flag. |
//  The source properties are: acceleration(ax,ay,az), potential, and mass-    |
//  density. If global softening is used, a NULL pointer may be used instead   |
//  of an array holding eps_i. The same applies to the arrays for density and  |
//  potential: if a NULL pointer is given, they will never be used.            |
//                                                                             |
//  The last 5 arguments of both constructors are identical:                   |
//  N:                  number of bodies / size of arrays.                     |
//  EPS:   if(EPS>=0):  globally fixed softening length                        |
//         if(EPS< 0):  use individual softening lengths                       |
//  THETA: if(THETA>0): theta = theta(M) with theta_min = THETA (Dehnen 2001b) |
//         if(THETA<0): theta = THETA = const                                  |
//         RECOMMENDED: THETA = 0.4                                            |
//  KERN:  newton     : no softening                                           |
//         f0,f1,f2,f3: Ferrers sphere softening kernel (Dehnen 2001a)         |
//         k0,k1,k2,k3: compensating softening kernel   (Dehnen 2001a)         |
//         p0,p1,p2,p3: Plummer(P0) and related (Dehnen & Teuben, in prep)     |
//  DMAX   maximum allowed tree depth (see 6.2 below)                          |
//         RECOMMENDED: DMAX = 50                                              |
//  The last three arguments may also be omitted, in which case the default    |
//  values are used.                                                           |
//                                                                             |
//  With the member functions                                                  |
//                                                                             |
    void reset_softening(const real,       // I:   softening length            |
			 const kern_type=f1)//[I]: softening kernel            |
                         const;            //                                  |
    void reset_opening  (const real) const;// I:   opening angle               |
//                                                                             |
//  it is possible to change, after construction, the softening length and     |
//  kernel as well as the opening criterion.                                   |
//                                                                             |
//                                                                             |
//  1.2 Destructor                                                             |
//  --------------                                                             |
//  The class falcON has a destructor that will be called implicitly, whenever |
//  the scope of an object of type falcON terminates. The destructor can also  |
//  be called explicitly, or via the delete command (if a tree was allocated   |
//  by the new command).                                                       |
//                                                                             |
    ~falcON();                             //                                  |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//                                                                             |
//  2 MEANING OF THE BODY FLAGS                                                |
//  ===========================                                                |
//                                                                             |
//  The flags are 2byte unsigned integers or plain integers with the following |
//  meaning:                                                                   |
//                                                                             |
//  bit  value     meaning                                                     |
//  ------------------------------------------------------------------------   |
//    1      1     this body is a sink, i.e. it wants update                   |
//    2      2     don't load this body into the tree, ignore it               |
//    3      4     this body is a SPH particle                                 |
//    4      8     this body is a sticky particle                              |
//  i>4   2^(i-1)  not used                                                    |
//                                                                             |
//  The default, flag=0, represents a plain body that is not sink, but still   |
//  source of gravity.                                                         |
//  The flag is obtained by setting the bits or, equivalently, adding the      |
//  values. The flags are not used by the constructor and will unfold their    |
//  effect only when the member functions, explained below, are used.          |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//                                                                             |
//  3 FORCE APPROXIMATION                                                      |
//  =====================                                                      |
//                                                                             |
//  3.1 Generating a tree structure                                            |
//  -------------------------------                                            |
//  In order to establish a hierarchical tree structure from all bodies that   |
//  are not flagged to be ignored, and subsequently pre-compute the centers    |
//  of mass and sizes of the cells, use                                        |
//                                                                             |
    void grow(const uint  = 6,             // [I]: Ncrit                       |
	      const vect* = 0);            // [I]: center of root              |
//                                                                             |
//  which creates a tree from scratch. Cells containing Ncrit or less bodies   |
//  will NOT be splitted. Experiments showed that for a full force calculation |
//  Ncrit = 6-8 is about optimal and results in a few % decrease in CPU time   |
//  consumption (compared to Ncrit=1) and about 20% reduction in memory.       |
//                                                                             |
//  You may instead also re-use an old tree structure and only re-compute the  |
//  centers of mass and sizes of the cells (since the bodies have moved), by   |
//  using                                                                      |
//                                                                             |
    void reuse();                          //                                  |
//                                                                             |
//  In this case, the logical linkage between cells and bodies is preserved    |
//  and no new tree structure is established. However, the code accounts for   |
//  the fact that the bodies might have moved. If the bodies have moved a lot, |
//  the sizes of the cells will be much larger than the physical size of the   |
//  associated boxes, and the tree traversal will be very inefficient.         |
//  However, if the bodies have moved only little, the force compuation is     |
//  hardly slowed down and re-using an old tree saves the costs for            |
//  establishing a new one. Note that reuse() does not allow for any change in |
//  the flags indicating whether or not a body shall be ignored: changing this |
//  flag will have an effect only at the next call of grow().                  |
//                     +---------------------------+                           |
//                     | USE THIS OPTION AREFULLY! |                           |
//                     +---------------------------+                           |
//  You have been warned!                                                      |
//                                                                             |
//                                                                             |
//  3.2 Approximating the forces                                               |
//  ----------------------------                                               |
//  Once a tree structure is established, you can compute the forces by        |
//                                                                             |
    void approximate_gravity();            //                                  |
//                                                                             |
//  which implements the pre-computation of the quadrupoles, as well as the    |
//  interaction and evaluation phase.                                          |
//                                                                             |
//  For test purposes, we also have a direct summation facility:               |
//                                                                             |
    void exact_gravity();                  //                                  |
//                                                                             |
//                                                                             |
//  3.3 Estimating mass- and number-density                                    |
//  ---------------------------------------                                    |
//  There is also the possibility to obtain a rough estimate of the mass-,     |
//  surface- or number density of bodies in the neighbourhood of every body    |
//  flagged as sink, via                                                       |
//                                                                             |
    void estimate_rho(const uint);         // I:  critical cell size           |
    void estimate_sd (const uint);         // I:  critical cell size           |
    void estimate_n  (const uint);         // I:  critical cell size           |
//                                                                             |
//  These estimates are simply the mean density within the smallest cells with |
//  more then NX (1st arg) bodies. Note, that when using test bodies (bodies   |
//  with zero or tiny mass), this guess for the mass density can have terrible |
//  errors. Moreover, when the tree has not been grow()n but simply reuse()ed, |
//  these estimate will not change. Be careful using these functions.          |
//  YOU HAVE BEEN WARNED.                                                      |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//                                                                             |
//  4 SEARCH FOR AND COUNTING OF NEIGHBOURS AND COLLISION PARTNERS             |
//  ==============================================================             |
//                                                                             |
//  Once a tree structure is established, you can also use it to create        |
//  interaction lists for neighbouring or colliding particles (useful for,     |
//  e.g., SPH and sticky-particle simulations). In the case of bodies (class   |
//  body) being used to communicate the data, this is provided by              |
//                                                                             |
    typedef uint elem_pair[2];             // element of an interaction list   |
    void make_iaction_list(elem_pair*,     // I:   interaction list            |
			   const uint,     // I:   allocated size of list      |
			   uint&,          // O:   actual size of list         |
			   const real=-1.);// [I]: time step tau               |
//                                                                             |
//  while in the case of array communication, you must use                     |
//                                                                             |
    void make_iaction_list(elem_pair*,     // I:   interaction list            |
			   const uint,     // I:   allocated size of list      |
			   uint&,          // O:   actual size of list         |
			   const areal*,   // I:   array with body sizes       |
			   const areal*[NDIM],//I: array with body velocities  |
			   const real=-1.);// [I]: time step tau               |
//                                                                             |
//  In case of overflow, i.e. if the number of pairs found exceeds the size    |
//  (2nd arg) of the list (1st arg), the routine issues a warning to stderr,   |
//  but does not truncate the search, rather it ceases copying pairs into the  |
//  interaction list. In this case, the value returned for the actual size     |
//  (3rd arg) exceeds the maximum size (2nd arg), but reflects the actual      |
//  number of interactions found.                                              |
//                                                                             |
//                                                                             |
//  4.1 Neighbour search (SPH support)                                         |
//  ----------------------------------                                         |
//  In order to make a list of all pairs {i,j} of bodies/indices for which     |
//                                                                             |
//       (1) both are flagged to be SPH particles,                             |
//  and  (2) at least one is flagged to be sink                                |
//  and  (3) | x_i - x_j | < max(size_i,size_j)                                |
//                                                                             |
//  use falcON::make_iaction_list() whereby omitting the time step (last arg). |
//  In the case of array communication, provide an array with sizes of the     |
//  bodies. It is the responsibility of the user to ensure that this array is  |
//  properly allocated (it is sufficient to have entries for all bodies        |
//  flagged as SPH particles).                                                 |
//                                                                             |
//                                                                             |
//  4.2 Collision partner search (for sticky particles)                        |
//  ---------------------------------------------------                        |
//  In order to make a list of all pairs {i,j} of bodies/indices for which     |
//                                                                             |
//       (1) both are flagged to be sticky particles,                          |
//  and  (2) at least one is flagged to be sink,                               |
//  and  (3) | (x_i+t*v_i)-(x_j+t*v_j) | < size_i+size_j  with t in [0,tau],   |
//                                                                             |
//  use falcON::make_iaction_list() with the time step (last arg) >= 0. In the |
//  case of array communication, provide arrays with sizes and velocity        |
//  components of the bodies. It is the responsibility of the user to ensure   |
//  that these arrays are properly allocated (it is sufficient to have entries |
//  for all bodies flagged as sticky).                                         |
//                                                                             |
//                                                                             |
//  4.3 SPH Neighbour counting                                                 |
//  --------------------------                                                 |
//  In order to count for each body flagged both as SPH particle and as sink   |
//  the number of other SPH particles within its size, use                     |
//                                                                             |
    void count_sph_neighbours();           // if bodies are used               |
    void count_sph_neighbours(             // if arrays are used               |
			      const areal*,// I:   array with sizes            |
			            areal*);//I/O: array with N_neighbour      |
//                                                                             |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//                                                                             |
//  5 OTHER FEATURES                                                           |
//  ================                                                           |
//                                                                             |
//  There are many further routines by which the user can obtain information   |
//  about the behaviour of the code.                                           |
//  For the number of bodies used by falcON, use                               |
//                                                                             |
    unsigned No_bodies_used() const;       //                                  |
//                                                                             |
//  For the number of cells in the tree, use                                   |
//                                                                             |
    unsigned No_cells_used() const;        //                                  |
//                                                                             |
//  For the number of cells (deleted) with only one active octant, use         |
//                                                                             |
    unsigned No_zombie_cells() const;      //                                  |
//                                                                             |
//  In order to dump (almost) all info of the bodies and cells to files, use   |
//                                                                             |
    void dump_nodes(const char* = 0,       // [I]: dump cells to this file?    |
		    const char* = 0) const;// [I]: dump bodies to this file?   |
//                                                                             |
//  To display some statistics, use                                            |
//                                                                             |
    void stats     (std::ostream&) const;  // I: output stream                 |
//                                                                             |
//  To write the opening method, use                                           |
//                                                                             |
    const char* opening_method() const;    //                                  |
//                                                                             |
//  To obtain the current type of softening (global/individual), use           |
//                                                                             |
    const soft_type& softening() const;    //                                  |
//                                                                             |
//  To obtain the currently used kernel, use                                   |
//                                                                             |
    const kern_type& kernel() const;       //                                  |
//                                                                             |
//  To obtain description of the currently used kernel, use                    |
//                                                                             |
    const char* describe_kernel() const;   //                                  |
    static const char* describe(const      //                                  |
				kern_type);//                                  |
//                                                                             |
//  To obtain the currently used (global) softening length, use                |
//                                                                             |
    const real& softening_length() const;  //                                  |
//                                                                             |
//  To obtain the currently used theta(mass)                                   |
//                                                                             |
    real theta(const real) const;          // I:   mass                        |
//                                                                             |
//  To obtain some interaction statistics, use                                 |
//                                                                             |
    uint BB_interactions() const;          // R:  # of body-body IAs           |
    uint MB_interactions() const;          // R:  # of many-body IAs           |
    uint CB_interactions() const;          // R:  # of cell-body IAs           |
    uint CC_interactions() const;          // R:  # of cell-cell IAs           |
    uint total_interactions() const;       // R:  # of all IAs                 |
//                                                                             |
//  To obtain some global tree properties, use                                 |
//                                                                             |
    const vect& root_center() const;       // R:  center of root               |
    const real& root_radius() const;       // R:  radius of root               |
    const uint& root_number() const;       // R:  # bodies in root             |
    const real& root_mass  () const;       // R:  mass  of root                |
    const uint& root_depth () const;       // R:  depth  of root               |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//                                                                             |
//  6 KNOWN BUGS AND PROBLEMS                                                  |
//  =========================                                                  |
//                                                                             |
//  6.1 Test-bodies are not possible                                           |
//  --------------------------------                                           |
//  A body that is loaded into the tree but has zero mass, will not acquire    |
//  any acceleration. This is because the code computes first the force = mass |
//  times acceleration (it is symmetric and hence better suited for the        |
//  computation of mutual interactions) and then divides by the mass to obtain |
//  the acceleration. The only possible work-around this problem is to set     |
//  the mass of potential test bodies much smaller than the masses of source   |
//  bodies. However, this must be done such that the gravity of test bodies    |
//  is everywhere neglible compared to that of the source bodies.              |
//  Note, however, that this work-around is wasteful: it computes the forces   |
//  generated by the test bodies.  (You have been warned!)                     |
//                                                                             |
//                                                                             |
//  6.2 Bodies at identical positions                                          |
//  ---------------------------------                                          |
//  The code cannot cope with more than Ncrit bodies at the same position      |
//  (within the floating point accuracy). This situation will lead to an       |
//  infinitely deep tree, i.e. the maximum allowed tree depth will be exceeded |
//  and the code aborts with an error message.                                 |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//                                                                             |
//  REFERENCES                                                                 |
//  ==========                                                                 |
//                                                                             |
//  Dehnen W., 2000, ApJ, 536, L39                                             |
//  Dehnen W., 2001a, MNRAS, 324, 273                                          |
//  Dehnen W., 2001b, submitted to J. Comp. Phys.                              |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//                                                                             |
//  A  NON-PUBLIC CODE AND CODE USED BY THE C AND FORTRAN SUPPORT              |
//  =============================================================              |
//                                                                             |
//  This is non-public and hence not commented at all.                         |
//                                                                             |
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  protected:
    const bodies* const& my_bodies() const { return BODIES; }
    const bodies*      & my_bodies()       { return BODIES; }
    const cell*          root()       const;
    cell*                root();
  protected:
    const uint& No_bodies       () const;
    uint        No_zombie_bodies() const;
  private:
    enum tree_state { unset=0, built=1, reused=2 };
    mutable tree_state     STATE;
    const bodies          *BODIES;
    tree<cell>            *TREE;
    uint                   Nb, Ncrit;
    const   int           *FL;
    const   areal         *M, *EP,**X;
    areal                 *PH,*RH,**A;
    mutable TaylorCoeffs  *T0;
    mutable real           EPS;
    mutable opening       *OPENING;
    mutable TreeStats     *STATS;
    mutable soft_type      SOFTENING;
    mutable kern_type      KERNEL;
    mutable indx           DMAX;

    falcON();                                     // not implemented            
    falcON(const falcON&);                        // not implemented            
    falcON& operator=(const falcON&);             // not implemented            
  };
  inline soft_type const &falcON::softening()        const { return SOFTENING; }
  inline kern_type const &falcON::kernel()           const { return KERNEL; }
  inline real      const &falcON::softening_length() const { return EPS; }
  inline const     char*  falcON::describe_kernel () const { return describe(KERNEL);}
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#endif // included_tree_h
