// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// nbdy.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// See the detailled comments for how to use this code. Some abbreviations:    |
//                                                                             |
// R:       return value                                                       |
// I:       function argument is input                                         |
// O:       function argument is output                                        |
// I/O:     function argument is input & output                                |
// [I]:     function argument is optional                                      |
//                                                                             |
// Note that in case of optional function arguments, one may only omit the     |
// last one (and the second but last if the last is already omitted ...).      |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// Note that it is often more convenient to use the wrapper provided by the    |
// classes defined in files yanc.h and ssgs.h                                  |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_nbdy_h
#define included_nbdy_h

#ifndef included_falcON__h
#include <falcON.h>
#endif

#ifndef included_pext_h
#include <tree/pext.h>
#endif

#ifndef included_step_h
#include <tree/step.h>
#endif
////////////////////////////////////////////////////////////////////////////////
namespace fasttree {
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::basic_nbody                                                
  //////////////////////////////////////////////////////////////////////////////
  class basic_nbody : 
    protected falcON
  {
    basic_nbody(const basic_nbody&);               // not implemented           
    basic_nbody& operator= (const basic_nbody&);   // not implemented           
    //--------------------------------------------------------------------------
    // types                                                                    
    //--------------------------------------------------------------------------
  public:
    enum soft_type {                               // type of softening         
      global_fixed        = 0,                     // globally time-constant    
      individual_fixed    = 1,                     // individual but time-const 
      individual_adaptive = 2                      // individual and adaptive   
    };
    //--------------------------------------------------------------------------
  protected:
    enum how_to_make_tree {
      grow  = 1,
      reuse = 2,
      free  = 4
    };
    //--------------------------------------------------------------------------
    typedef real tensor[NDIM][NDIM];               // not necessarily symmetric 
    //--------------------------------------------------------------------------
    // data                                                                     
    //--------------------------------------------------------------------------
  protected:
    const bodies  *BODIES;
    const potext  *PEX;
    soft_type      SOFTENING;                      // type of softening method  
    real           TINI;                           // initial simulation time   
  private:
    uint           NSOFT;
    uint           NREF;
    const uint     NREUSE;
    uint           N_REUSED;
    const uint     NCRIT;                          // max # bodies in cell      
    real           EMAX;                           // max softening length      
    mutable real   M,Ktot,Utot,TU;                 // mass, kin & pot E, vir rat
    mutable amom   L;                              // total angular momentum    
    mutable tensor KT,WT;                          // kin & pot energy, AM tens 
    mutable vect   CMX,CMV;                        // center of mass pos & vel  
  protected:
    mutable bool   DIAG;                           // diagnose up-to-date ?     
    mutable real   CPU_BUILD,                      // time for falcON::grow etc 
                   CPU_GRAV,                       // time for gravity etc      
                   CPU_PEX,                        // time for ext potential    
                   CPU_PREDICT,                    // time for advancing Xs     
                   CPU_DENS,                       // time for density estimates
                   CPU_EPSI,                       // time for setting eps_i    
                   CPU_STEP,                       // total time for a longstep 
                   CPU_TOTAL;                      // total time so far         
    //--------------------------------------------------------------------------
    // protected methods                                                        
    //--------------------------------------------------------------------------
  protected:
    void diagnose             () const;
    void update_diagnostics   () const;
    real trace                (const tensor&) const;
    void set_tree             (const vect* =0, const how_to_make_tree=free);
    void set_forces           ();
    void reset_cpus           () const;
    void setup_eps            (body&) const;
    void set_softening_lengths();
    void estimate_mass_density();
    //--------------------------------------------------------------------------
    // abstract methods                                                         
    //--------------------------------------------------------------------------
  public:
    virtual const real& time() const=0;            // actual time               
    virtual void full_step  (const vect* =0)=0;    // one full time step        
    virtual void long_stats (std::ostream&) const=0;
    virtual void short_stats(std::ostream&) const=0;
    virtual void short_stat_head (std::ostream&) const=0;
    virtual void short_stat_line (std::ostream&) const=0; 
    //--------------------------------------------------------------------------
    // public methods                                                           
    //--------------------------------------------------------------------------
    void estimate_mass_densities();                // estimate every bodies rho 
    void estimate_surf_densities();                // estimate every bodies SD  
    const real& kin_energy      () const;          // T   = kinetic energy      
    const real& pot_energy      () const;          // U   = potential energy    
    const real  total_energy    () const;          // T+U = total energy        
    const real& virial_ratio    () const;          // -T/U                      
    const amom& total_angmom    () const;          // total angular momentum    
    const vect& total_momentum  () const;          // total momentum            
    const vect& center_of_mass  () const;          // center of mass            
    const real& kin_energy      (const indx, const indx) const; // K_ij         
    const real& pot_energy      (const indx, const indx) const; // W_ij         
    const real  as_angmom       (const indx, const indx) const; // anti-sym AM  
    const real& cpu_predict     () const { return CPU_PREDICT; }
    const real& cpu_build       () const { return CPU_BUILD; }
    const real& cpu_tree_force  () const { return CPU_GRAV; }
    const real& cpu_ext_force   () const { return CPU_PEX; }
    const real  cpu_force       () const { return CPU_GRAV+CPU_PEX; }
    const real& cpu_density     () const { return CPU_DENS; }
    const real& cpu_softening   () const { return CPU_EPSI; }
    const real& cpu_longstep    () const { return CPU_STEP; }
    const real& cpu_total       () const { return CPU_TOTAL; }
    //--------------------------------------------------------------------------
    void reset_softening(                          // resets softening params   
			 const soft_type,          // I: type of softening      
			 const unsigned,           // I: Nsoft                  
			 const real,               // I: eps                    
			 const kern_type);         // I: softening kernel       
    //--------------------------------------------------------------------------
    void reset_opening(const real th) const { falcON::reset_opening(th); }
    //--------------------------------------------------------------------------
    const real t_ini() const { return TINI; }
    //--------------------------------------------------------------------------
    basic_nbody(const bodies    *,                 // I: bodies                 
		const uint       ,                 // I: N_soft                 
		const uint       ,                 // I: N_ref                  
		const real       ,                 // I: eps/eps_max            
		const real       ,                 // I: initial time           
		const soft_type  = global_fixed,   //[I: softening type]        
		const real       = half,           //[I: tolerance parameter]   
		const uint       = 0,              //[I: N_reuse]               
		const uint       = 6,              //[I: N_crit]                
		const kern_type  = p1,             //[I: softening kernel]      
		const potext   * = 0);             //[I: P_ex]                  
    //--------------------------------------------------------------------------
    virtual ~basic_nbody() {}
  };
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::LeapFrogCode                                               
  //////////////////////////////////////////////////////////////////////////////
  class LeapFrogCode :
    public    basic_nbody,
    protected LeapFrog
  {
    LeapFrogCode(const LeapFrogCode&);              // not implemented          
    LeapFrogCode& operator= (const LeapFrogCode&);  // not implemented          
    //--------------------------------------------------------------------------
    // public methods                                                           
    //--------------------------------------------------------------------------
  public:
    LeapFrogCode(const bodies    *,                // I: bodies                 
		 const uint       ,                // I: N_soft                 
		 const real       ,                // I: eps/eps_max            
		 const real       ,                // I: initial time           
		 const int        ,                // I: h0                     
		 const uint       = 0,             //[I: N_reuse]               
		 const uint       = 16u,           //[I: N_ref]                 
		 const vect     * = 0,             //[I: center of root]        
		 const soft_type  = global_fixed,  //[I: softening type]        
		 const real       = half,          //[I: tolerance parameter]   
		 const uint       = 6,             //[I: N_crit]                
		 const kern_type  = p1,            //[I: softening kernel]      
		 const potext   * = 0);            //[I: P_ex]                  
    //--------------------------------------------------------------------------
    void full_step      (const vect* = 0);         //[I: center of root ]       
    const real& time() const { return LeapFrog::time(); }
    void long_stats      (std::ostream&) const;    // statistic of long_step    
    void short_stats     (std::ostream&) const;    // statistic of long_step    
    void short_stat_head (std::ostream&) const;    // statistic of long_step    
    void short_stat_line (std::ostream&) const;    // line of proper size       
  };
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::BlockStepCode                                              
  //////////////////////////////////////////////////////////////////////////////
  class BlockStepCode :
    public    basic_nbody,
    protected BlockStep
  {
    BlockStepCode(const BlockStepCode&);            // not implemented          
    BlockStepCode& operator= (const BlockStepCode&);// not implemented          
    //--------------------------------------------------------------------------
    // types                                                                    
    //--------------------------------------------------------------------------
  public:
    enum step_crit {
      dacc = 0,
      rhos = 1
    };
    //--------------------------------------------------------------------------
    // data                                                                     
    //--------------------------------------------------------------------------
  private:
    step_crit      MYSCHEME;                       // time stepping scheme      
    const uint     NREF;
    real           FAC,TQ;                         // time-step contolling facs 
    //--------------------------------------------------------------------------
    // private methods                                                          
    //--------------------------------------------------------------------------
    void prepare(const int   ,                     // I:   h0                   
		 const indx  ,                     // I:   N_stps               
		 const vect* =0);                  //[I: center of root]        
    //--------------------------------------------------------------------------
    void      set_pposs();
    //--------------------------------------------------------------------------
    // routines to use with rhos scheme                                         
    //--------------------------------------------------------------------------
    bool      test_for_longer(const indx, const body&) const;
    bool      test_for_shorter(const indx, const body&) const;
    void      smallest_step_rhos(const indx, const vect*,
				 const how_to_make_tree=free);
    //--------------------------------------------------------------------------
    // routines to use with dacc scheme                                         
    //--------------------------------------------------------------------------
    bool      may_go_longer (const body&) const;
    bool      step_too_long (const body&) const;
    too_short step_too_short(const body&) const;
    void      set_sink_flags(const indx);
    void      smallest_step_dacc(const indx, const vect*,
				 const how_to_make_tree=free);
    //--------------------------------------------------------------------------
    // public methods                                                           
    //--------------------------------------------------------------------------
  public:
    BlockStepCode(const bodies    *,               // I: bodies                 
		  const uint       ,               // I: N_soft                 
		  const real       ,               // I: eps/eps_max            
		  const real       ,               // I: initial time           
		  const int        ,               // I: h0                     
		  const indx       ,               // I: N_steps                
		  const uint       ,               // I: N_ref                  
		  const real       ,               // I: factor for stepping    
		  const vect     * = 0,            //[I: center of root]        
		  const real       = two,          //[I: tol^2]                 
		  const soft_type  = global_fixed, //[I: softening type]        
		  const real       = half,         //[I: tolerance parameter]   
		  const uint       = 6,            //[I: N_crit]                
		  const kern_type  = p1,           //[I: softening kernel]      
		  const potext   * = 0);           //[I: P_ex]                  
    //--------------------------------------------------------------------------
    void reset_stepping (const real);              // I: f                      
    void full_step      (const vect* = 0);         //[I: center of root ]       
    const step_crit& stepping()const{ return MYSCHEME; }
    const real& time() const { return BlockStep::time(); }
    void dump_steps(std::ostream& to) const { BlockStep::dump(to); }
    void long_stats      (std::ostream&) const;    // statistic of long_step    
    void short_stats     (std::ostream&) const;    // statistic of long_step    
    void short_stat_head (std::ostream&) const;    // statistic of long_step    
    void short_stat_line (std::ostream&) const;    // line of proper size       
  };
//-----------------------------------------------------------------------------+
//                                                                             |
//  ON THE TIME-STEPPING SCHEMES                                               |
//  ============================                                               |
//                                                                             |
//  There are two possible options, the nearly time symmetric and the exactly  |
//  time symmetric scheme of Dehnen (2001b). Both use a block step scheme with |
//  hierarchically nested time steps that differ by a factor of two between    |
//  two levels of the hierarchy.                                               |
//                                                                             |
//  2.1 The nearly time symmetric scheme                                       |
//  ------------------------------------                                       |
//  This scheme is based on an ESTIMATE, made from the first half of each      |
//  step, for the relative change in each bodies acceleration. The idea is to  |
//  limit this change to be less than f:                                       |
//                                                                             |
//         |delta acc| < f |acc|                                               |
//                                                                             |
//  Besides computing the acceleration at the beginning and end of each step,  |
//  which is necessary for the leap-frog time integrator, the code also        |
//  computes it in the MIDDLE, to (i) check the criterion and (ii) to use it   |
//  in the time integration in case a shorter step is required.                |
//  There is also another parameter, tol, which can be used to reduce the      |
//  number of these extra computations of acceleration in the middle of each   |
//  step: Only when                                                            |
//                                                                             |
//         |delta acc| < f * tol |acc|                                         |
//                                                                             |
//  will the criterion actually be checked. Thus tol=2 results in the strict   |
//  case with checking every time, while tol=1 means no checking in the middle |
//  of each step. The default value, tol=sqrt[2], controls half the bodies.    |
//                                                                             |
//  2.2 The exactly time symmetric scheme                                      |
//  -------------------------------------                                      |
//  This scheme uses the mass density rho to assign time steps, according to   |
//                                                                             |
//         tau < f / rho                                                       |
//                                                                             |
//  where rho is estimated in the middle of each step (which makes the scheme  |
//  time symmetric) as the mean density of the smalles cell that (i) contains  |
//  the body in question and (ii) at least Nref bodies in total.               |
//  With this scheme, no acceleration is computed without being used in the    |
//  time integration.                                                          |
//                                                                             |
//-----------------------------------------------------------------------------+
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_nbdy_h
