// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// iact.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// template<class A, class B> struct iaction                                   |
// template<class A, class B> class iastack                                    |
// template<class INTERACTOR, class CELL, class SOUL> void interact()          |
// template<class INTERACTOR, class CELL, class SOUL> class interaction        |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_iact_h
#define included_iact_h

namespace fasttree {
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::iaction<A,B>                                               
  //////////////////////////////////////////////////////////////////////////////
  template<class A, class B>
  struct iaction
  // holds a pair of objects, very simple                                       
  {
    A     fst;                                     // first object              
    B     snd;                                     // second object             
    iaction (A a=0, B b=0) :                       // constructor               
      fst (a),                                     //   set first               
      snd(b) {}                                    //   set second              
    void set(A a, B b)                             // set pair                  
    {                                              // >                         
      fst =a;                                      //   set first               
      snd=b;                                       //   set second              
    }                                              // <                         
    iaction& operator= (const iaction&I)           // assignment                
    {                                              // >                         
      fst =I.fst;                                  //   set first               
      snd=I.snd;                                   //   set second              
      return *this;                                //   return                  
    }                                              // <                         
  };
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::iastack<A,B>                                               
  //////////////////////////////////////////////////////////////////////////////
  template<class A, class B>
  class iastack
  // manages a stack of iaction<>s, very simple                                 
  {
  private:
    typedef iaction<A,B>    iact;                  // type of stack objects     
    iact    *IA, *pi;                              // first & active element    
  public:
    iastack (const unsigned int M) :               // constructor               
      IA(new iact[M]),                             //   allocate memory         
      pi(IA-1) {}                                  //   set pter to active      
    ~iastack () { delete[] IA; }                   // destructor: deallocate    
    bool is_empty() const   { return pi<IA;}       // if stack empty?           
    iact pop     ()         { return *(pi--); }    // give last active: pop     
    void push    (A a, B b) { (++pi)->set(a,b); }  // add element:      push    
  };
  //////////////////////////////////////////////////////////////////////////////
  // functions fasttree::MutualInteract<I,A,B>()                                
  //////////////////////////////////////////////////////////////////////////////
#define LoopCKids(C,A)    LoopCellKids(typename CELL,C,A)
#define LoopSKids(C,A)    LoopSoulKids(typename CELL,C,A)
#define LoopCPairs(C,A,B) for(register typename CELL::cell_child	\
			      B=A; B!=C->end_cell_children(); ++B)
#define LoopSPairs(C,A,B) LoopSKids(C,A)				\
                            for(register typename CELL::soul_child	\
			        B=A+1; B!=C->end_soul_children(); ++B)
  //----------------------------------------------------------------------------
  template<class INTERACTOR, class CELL, class SOUL>
  inline void MutualInteract(      CELL      *I1,  // I: interacting cell 1     
		             const uint       D1,  // I: depth of cell 1        
		                   CELL      *I2,  // I: interacting cell 2     
		             const uint       D2,  // I: depth of cell 2        
		             const INTERACTOR*K)   // I: interactor             
  {
    const unsigned int  NN = (I1==I2)?             // self interaction ?        
                        NSUB*(NSUB+1)*D1/2         // max size of stacks        
                      : NSUB*(D1+D2);              // max size of stacks        
    register iastack<CELL*,CELL*> CCS(NN);         // stack for C-C interactions
    register iastack<CELL*,SOUL*> CBS(NN);         // stack for C-B interactions
    register iaction<CELL*,CELL*> cc;              // current C-C interaction   
    register iaction<CELL*,SOUL*> cb;              // current C-B interaction   
    CCS.push(I1,I2);                               // push I1-I2 onto CC        
    while(!CCS.is_empty()) {                       // while CC non-empty  >     
      cc = CCS.pop();                              //   pop new C-C interaction 
      if(cc.fst == cc.snd) {                       //   IF self iinteraction   >
	if(K->interact(cc.fst)) continue;          //     IF success: next      
	LoopCKids(cc.fst,c1) {                     //     ELSE: split     >     
	  LoopSKids (cc.fst,s2)    CBS.push(c1,s2);//       push sub C-B        
	  LoopCPairs(cc.fst,c1,c2) CCS.push(c1,c2);//       push sub C-C        
	}                                          //     <                     
	LoopSPairs(cc.fst,s1,s2)K->interact(s1,s2);//     do sub B-B            
      } else {                                     //   < ELSE mutual iaction  >
	if(K->interact(cc.fst,cc.snd)) continue;   //     success? -> next      
	if(K->split_first(cc.fst,cc.snd)) {        //     IF split 1st    >     
	  LoopSKids(cc.fst,s2) CBS.push(cc.snd,s2);//       push sub C-B        
	  LoopCKids(cc.fst,c2) CCS.push(c2,cc.snd);//       push sub C-C        
	} else {                                   //     < ELSE split 2nd>     
	  LoopSKids(cc.snd,s2) CBS.push(cc.fst,s2);//       push sub C-B        
	  LoopCKids(cc.snd,c2) CCS.push(cc.fst,c2);//       push sub C-C        
	}                                          //     <                     
      }                                            //   <                       
      while(!CBS.is_empty()) {                     //   while CB non-empty >    
	cb = CBS.pop();                            //     pop new C-B iaction   
	if(K->interact(cb.fst,cb.snd)) continue;   //     success? -> next      
	LoopCKids(cb.fst,c1)CBS.push(c1,cb.snd);   //     push sub C-B          
	LoopSKids(cb.fst,s2)K->interact(cb.snd,s2);//     do sub B-B            
      }                                            //   <                       
    }                                              // <                         
  }
  //----------------------------------------------------------------------------
#undef LoopCKids
#undef LoopSKids
#undef LoopCPairs
#undef LoopSPairs
  //----------------------------------------------------------------------------
  template<class INTERACTOR, class CELL, class SOUL>
  inline void MutualInteract(      CELL      *I,   // I: self-iacting cell      
		             const uint       D,   // I: depth of cell          
		             const INTERACTOR*K)   // I: interactor             
  {
    MutualInteract<INTERACTOR,CELL,SOUL>(I,D,I,D,K);
  }
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_iact_h
