// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// inln.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 1994-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// various useful inline functions, often templates                            |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_inln_h
#define included_inln_h

#ifndef included_algorithm
#include <algorithm>
#define included_algorithm
#endif

#ifndef included_iostream
#include <iostream>
#define included_iostream
#endif

#ifndef included_fstream
#include <fstream>
#define included_fstream
#endif

#ifndef included_string
#include <string>
#define included_string
#endif

//------------------------------------------------------------------------------
namespace fasttree {
  template<class S> inline
  S    abs     (const S x)              { return (x<0)? -x : x; }
  //----------------------------------------------------------------------------
  template<class S> inline
  int  sign    (const S x)              { return (x<0)? -1:((x>0)? 1:0 ); }
  //----------------------------------------------------------------------------
  template<class S> inline
  S    sign    (const S x, const S s)   { return ( s>0 )?  abs(x) : -abs(x); }
  //----------------------------------------------------------------------------
  template<class S> inline
  S    min     (const S x, const S y)   { return (x<y)? x : y; }
  //----------------------------------------------------------------------------
  template<class S> inline
  S    max     (const S x, const S y)   { return (x>y)? x : y; }
  //----------------------------------------------------------------------------
  template<class S> inline
  S    mod     (const S x, const S y)   { return x-y*int(x/y); }
  //----------------------------------------------------------------------------
  template<class S> inline
  S    twice   (const S x)              { return x+x; }
  //----------------------------------------------------------------------------
  template<class S> inline
  S    trice   (const S x)              { return x+x+x; }
  //----------------------------------------------------------------------------
  template<class S> inline
  S    square  (const S x)              { return x*x; }
  //----------------------------------------------------------------------------
  template<class S> inline
  S    cube    (const S x)              { return x*x*x; }
  //----------------------------------------------------------------------------
  template<class S> inline
  void swap    (S&a, S&b)               { register S t=a; a=b; b=t; }
  //----------------------------------------------------------------------------
  template<class S> inline
  S    pow_uint(const S x, const unsigned int n)
  {
    if(n==0) return S(1);
    register S z=x, y=(n%2)? x : S(1);
    for(register unsigned int i=n>>1; i; i>>=1) { z*=z; if(i%2) y*=z; }
    return y;
  }
  //----------------------------------------------------------------------------
  template<class S> inline
  S    pow_int (const S x, const int i)
  {
    if(i>=0) return pow_uint(x,i);
    if(x==0) { std::cerr<<"pow_int(): negative power of zero"; std::exit(1); }
    return pow_uint(S(1)/x,-i);
  }
  //----------------------------------------------------------------------------
  template<class S> inline
  void update_max(S&x, const S y) { if(y>x) x=y; }
  //----------------------------------------------------------------------------
  template<class S> inline
  void update_min(S&x, const S y) { if(y<x) x=y; }
  //----------------------------------------------------------------------------
  template<class X>
  inline char* negspace(const X x)
  {
    return (x<X(0))? " " : "  ";
  }
  //----------------------------------------------------------------------------
  inline int open(std::ofstream& S, const char* file,
		  std::ios::openmode mode = std::ios::out | std::ios::trunc)
    // 		int mode= ios::out)
  {
    S.open(file,mode);
    if(! S.is_open() ) {
      std::cerr<<" cannot open file \""<<file<<"\"";
      return 0;
    }
    return 1;
  }
  //----------------------------------------------------------------------------
  inline int open(std::ofstream& S, const std::string file,
		  std::ios::openmode mode=std::ios::out)
  {
    return open(S,file.c_str(),mode);
  }

  //----------------------------------------------------------------------------
  inline int open_to_append(std::ofstream& S, const char* file)
  {
    S.open(file,std::ios::out | std::ios::app);
    if(! S.is_open() ) S.open(file,std::ios::out);
    if(! S.is_open() ) {
      std::cerr<<" cannot open file \""<<file<<"\"";
      return 0;
    }
    return 1;
  }

  //----------------------------------------------------------------------------
  inline int open_to_append(std::ofstream& S, const std::string file)
  {
    return open_to_append(S,file.c_str());
  }

  //----------------------------------------------------------------------------
  inline int open(std::ifstream& S, const char* file,
		  std::ios::openmode mode=std::ios::in)
  {
    S.open(file,mode);
    if(! S.is_open() ) {
      std::cerr<<" cannot open file \""<<file<<"\"";
      return 0;
    }
    return 1;
  }

  //----------------------------------------------------------------------------
  inline int open(std::ifstream& S, const std::string file,
		  std::ios::openmode mode=std::ios::in)
  {
    return open(S,file.c_str(),mode);
  }

  //----------------------------------------------------------------------------
  inline int open(std::fstream& S, const char* file, std::ios::openmode mode)
  {
    S.open(file,mode);
    if(! S.is_open() ) {
      std::cerr<<" cannot open file \""<<file<<"\"";
      return 0;
    }
    return 1;
  }

  //----------------------------------------------------------------------------
  inline int open(std::fstream& S, const std::string file,
		  std::ios::openmode mode)
  {
    return open(S,file.c_str(),mode);
  }

  //----------------------------------------------------------------------------
  inline void SwallowRestofLine(std::istream& from)
  {
    char c;
    do from.get(c); while( from.good() && c !='\n');
  }
  //----------------------------------------------------------------------------
  inline const char* stndrdth(const int i)
  {
    register int ia= (i<0)? -i:i;
    switch( ia % 100 ) {
    case 11: 
    case 12: 
    case 13: return "th";
    default:
      switch( ia % 10 ) {
      case 1:  return "st";
      case 2:  return "nd"; 
      case 3:  return "rd";
      default: return "th";
      }   
    }
  }
  //----------------------------------------------------------------------------
  template<typename REAL, unsigned N>
  std::ostream& write_array(std::ostream&s, const REAL* x) {
    s << x[0];
    for(register int i=1; i!=N; ++i) s<<" "<<x[i];
    return s;
  }
  //----------------------------------------------------------------------------
  template<typename REAL, unsigned N>
  std::istream& read_array(std::istream&s, REAL* x) {
    register REAL y[N];
    char c=0;
    s >> c;
    if(c == '(') {
      for(register int i=0; i!=N; ++i) s >> y[i];
      s >> c;
      if(c != ')') s.clear(std::ios::badbit);
    } else {
      s.putback(c);
      for(register int i=0; i!=N; ++i) s >> y[i];
    }
    for(register int i=0; i!=N; ++i) x[i] = y[i];
    return s;
  }
}
//------------------------------------------------------------------------------
#endif // included_inln_h
