// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// soul.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class soul                fully inline                                      |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_soul_h
#define included_soul_h

#ifndef included_auxx_h
#include <tree/auxx.h>
#endif

#ifndef included_flag_h
#include <tree/flag.h>
#endif

#ifndef included_bsoul_h
#include <tree/bsoul.h>
#endif

////////////////////////////////////////////////////////////////////////////////
namespace fasttree {
  class bodies;                                    // forward declaration       
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::soul                                                       
  //////////////////////////////////////////////////////////////////////////////
  class soul :
    public flagged_soul<bodies>
  {
    soul           (const soul&);                  // not implemented           
    soul& operator=(const soul&);                  // not implemented           
    //--------------------------------------------------------------------------
    // data of class soul                                                       
    //--------------------------------------------------------------------------
  private:
    real        MASS;                              // mass of body              
    vect        ACC;                               // acceleration              
    real        POT;                               // potential                 
    real        RHO;                               // mass density              
    real        EPS;                               // individual softening      
    //--------------------------------------------------------------------------
    // constructor                                                              
    //--------------------------------------------------------------------------
  public:
    soul() {}
    //--------------------------------------------------------------------------
    // simple manipulations for use with bodies                                 
    //--------------------------------------------------------------------------
    void set_up         (const bodies*);           // set cofm, mass from body  
    void link_and_set_up(const bodies*,
			 const uint&);             // set my body, cofm & mass  
    void reset_grav     (const bodies*);           // copy sink flag, reset grav
    //--------------------------------------------------------------------------
    // simple manipulations for use with arrays                                 
    //--------------------------------------------------------------------------
    void set_up         (const areal*,             // set cofm, mass from arrays
			 const areal*[NDIM]);      // I: arrays with mass,cofm  
    void link_and_set_up(const vector_type,        // set my element & cofm     
			 const uint);              // I: cofm , element         
    void set_mass       (const areal*);            // set mass from array       
    void reset_grav     (const flag_type*);        // copy sink flag, reset grav
    void reset_grav_mass(const flag_type*,         // copy sink flag, reset grav
			 const areal*);            //                 and mass  
    void reset_grav     (const flag_type*,         // copy sink flag, reset grav
			 const areal*);            //                 and eps   
    void reset_grav_mass(const flag_type*,         // copy sink flag, reset grav
			 const areal*,             //                 and mass  
			 const areal*);            //                 and eps   
    void update_grav    (areal*[NDIM],areal*);     // copy acc,pot into arrays  
    void update_dens    (areal*);                  // copy rho into array       
    //--------------------------------------------------------------------------
    // simple manipulations for general use                                     
    //--------------------------------------------------------------------------
    void normalize_grav ();                        // acc,pot     /= mass       
    //--------------------------------------------------------------------------
    // non-const data access via members                                        
    //--------------------------------------------------------------------------
    vect      &acc()  { return ACC; }
    real      &pot()  { return POT; }
    real      &rho()  { return RHO; }
    //--------------------------------------------------------------------------
    // const access via friends                                                 
    //--------------------------------------------------------------------------
#define CSCS const soul* const &S
    friend vector_type const &cofm     (CSCS) { return S->cofm(); }
    friend real        const &mass     (CSCS) { return S->MASS; }
    friend real        const &eps      (CSCS) { return S->EPS; }
    friend vect        const &acc      (CSCS) { return S->ACC; }
    friend real        const &pot      (CSCS) { return S->POT; }
    friend real        const &rho      (CSCS) { return S->RHO; }
    friend bool               is_source(CSCS) { return S->MASS != zero; }
#undef CSCS
    //--------------------------------------------------------------------------
    // other const and static methods                                           
    //--------------------------------------------------------------------------
    static void dump_head(ostream&);
    void        dump     (ostream&, const soul*) const;
  };
}
////////////////////////////////////////////////////////////////////////////////
#ifndef included_body_h
#include <body.h>
#endif
////////////////////////////////////////////////////////////////////////////////
namespace fasttree {
  //----------------------------------------------------------------------------
  // simple manipulations for use with bodies                                   
  //----------------------------------------------------------------------------
  inline void soul::set_up(const bodies*B) {
    MASS   = B->mass(mybody());
    cofm() = B->pos (mybody());
  }
  //----------------------------------------------------------------------------
  inline void soul::link_and_set_up(const bodies*B, const uint&i) {
    link  (i);
    set_up(B);
  }
  //----------------------------------------------------------------------------
  inline void soul::reset_grav(const bodies*B) {
    copy_sink_flag(B);
    EPS = B->eps(mybody());
    ACC = zero;
    POT = zero;
  }
  //----------------------------------------------------------------------------
  // simple manipulations for use with arrays                                   
  //----------------------------------------------------------------------------
  inline void soul::set_up(const areal*m, const areal*x[NDIM]) {
    MASS     = m[mybody()];
    for(register int d=0; d<NDIM; ++d) cofm()[d] = x[d][mybody()];
  }
  //----------------------------------------------------------------------------
  inline void soul::link_and_set_up(const vector_type x, const uint i) {
    link(i);
    cofm()=x;
  }
  //----------------------------------------------------------------------------
  inline void soul::set_mass(const areal*m) {
    MASS = m[mybody()];
  }
  //----------------------------------------------------------------------------
  inline void soul::update_grav(areal*a[NDIM], areal*p) {
    if(p) p[mybody()] = POT;
    for(register int d=0; d<NDIM; ++d) a[d][mybody()] = ACC[d];
  }
  //----------------------------------------------------------------------------
  inline void soul::update_dens(areal*rh) {
    rh[mybody()] = RHO;
  }
  //----------------------------------------------------------------------------
  inline void soul::reset_grav(const flag_type*F) { 
    copy_sink_flag(F);
    ACC = zero;
    POT = zero;
  }
  //----------------------------------------------------------------------------
  inline void soul::reset_grav(const flag_type*F, const areal* E) { 
    copy_sink_flag(F);
    EPS  = E[mybody()];
    ACC  = zero;
    POT  = zero;
  }
  //----------------------------------------------------------------------------
  inline void soul::reset_grav_mass(const flag_type*F,
				    const areal* M) { 
    copy_sink_flag(F);
    MASS = M[mybody()];
    ACC  = zero;
    POT  = zero;
  }
  //----------------------------------------------------------------------------
  inline void soul::reset_grav_mass(const flag_type*F,
				    const areal* M, const areal* E) { 
    copy_sink_flag(F);
    MASS = M[mybody()];
    EPS  = E[mybody()];
    ACC  = zero;
    POT  = zero;
  }
  //----------------------------------------------------------------------------
  // simple manipulations for general use                                       
  //----------------------------------------------------------------------------
  inline void soul::normalize_grav() {
    if(MASS>zero) {
      register real im = one/MASS;
      ACC *= im; POT *= im;
    }
  }
  //----------------------------------------------------------------------------
  // other const and static methods                                             
  //----------------------------------------------------------------------------
  inline void soul::dump_head(ostream& o) {
    o<<"     #";
    flagged_soul<bodies>::dump_head(o);
    o<<"              mass";
  }
  //----------------------------------------------------------------------------
  inline void soul::dump(ostream&o, const soul*S0) const
  {
    o<<" "<<setw(5)<<int(this-S0);
    flagged_soul<bodies>::dump(o);
    o<<" "<<setw(8)<<MASS;
  }
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_soul_h
