// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// tayl.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class TaylorCoeffs             fully inline                                 |
// class TaylorGrav               see tayl.cc                                  |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_tayl_h
#define included_tayl_h

#ifndef included_auxx_h
#include <tree/auxx.h>
#endif

#ifndef included_soul_h
#include <tree/soul.h>
#endif

////////////////////////////////////////////////////////////////////////////////

namespace fasttree {

////////////////////////////////////////////////////////////////////////////////
// useful constants and macros

#ifdef TWODIMENSIONAL
  const indx N_C2=3, N_C3=6,  N_G=10;
#else
  const indx N_C2=4, N_C3=10, N_G=20;
#endif // TWODIMENSIONAL

#define FI for(register indx i=0; i!=N_G; i++)
#define op operator

////////////////////////////////////////////////////////////////////////////////

  class soul;                               // forward declaration              

  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::TaylorCoeffs                                               
  //////////////////////////////////////////////////////////////////////////////
  class TaylorCoeffs {                      // sink properties of a cell
  private:
    mutable real         G[N_G];            //   accumulated gravity
  protected:
    void reset  ()                      { FI G[i]=zero; }
    // constructors
  public:
    TaylorCoeffs()                      { reset(); }
  protected:
    TaylorCoeffs(const TaylorCoeffs& T) { FI G[i]=T.G[i]; }
    // some manipulators
    TaylorCoeffs& shift_by    (const vect&);
    void          extract_grav(const vect&, soul*) const;
    void          addmul      (const TaylorCoeffs& T, const real x)
                                        { FI G[i] += T.G[i]*x; }
  public:
    // assign opertators
    TaylorCoeffs& op =(const TaylorCoeffs& T) { FI G[i] =T.G[i]; return *this; }
    TaylorCoeffs& op+=(const TaylorCoeffs& T) { FI G[i]+=T.G[i]; return *this; }
    TaylorCoeffs& op*=(const real x)          { FI G[i]*=x;      return *this; }
    // non-const data access via members
    real&       c0()                          { return G[0]; }
    evec        c1()                          { return evec(G+1); }
    emat        c2()                          { return emat(G+N_C2); }
    eten        c3()                          { return eten(G+N_C3); }
    // const data access via friends
    friend const real& c0(const TaylorCoeffs*T) { return T->G[0]; }
    friend const evec  c1(const TaylorCoeffs*T) { return evec(T->G+1); }
    friend const emat  c2(const TaylorCoeffs*T) { return emat(T->G+N_C2); }
    friend const eten  c3(const TaylorCoeffs*T) { return eten(T->G+N_C3); }
    // boolean informations via friends
    friend bool is_empty (const TaylorCoeffs*T) { return T->G[0]==zero; }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::TaylorGrav                                                 
  //////////////////////////////////////////////////////////////////////////////
  class TaylorGrav : public TaylorCoeffs {
  private:
    vect X0;
  public:
    TaylorGrav(const vect X=zero) : X0(X) {}
    TaylorGrav(const evec X) : X0(X) {}
    TaylorGrav(const TaylorGrav& T) : TaylorCoeffs(T), X0(T.X0) {}
    TaylorGrav(const TaylorCoeffs& T, const vect X=zero)
      : TaylorCoeffs(T), X0(X) {}
    void  reset (const vect X=zero)     { TaylorCoeffs::reset(); X0=X; }
    vect& center()                      { return X0; }
    const vect& center() const                { return X0; }
    TaylorGrav& op =(const TaylorGrav& T) { TaylorCoeffs::op= (T); X0=T.X0;
    return *this; }
    TaylorGrav& op*=(const real x)
    { TaylorCoeffs::op*=(x); return *this; }
    TaylorGrav& shift_to(const vect& X)
    { shift_by(X-X0); X0=X; return *this; }
    TaylorGrav& shift_to(const evec& X)
    { shift_by(X-X0); X0=X; return *this; }
    void extract_grav(soul*S) const
    { TaylorCoeffs::extract_grav(cofm(S)-X0,S); }
    void shift_and_add(const vect& X, const TaylorCoeffs* T, const real& iM)
    { if(!(is_empty(T))) { shift_to(X); addmul(*T,iM); } }
    void shift_and_add(const evec& X, const TaylorCoeffs* T, const real& iM)
    { if(!(is_empty(T))) { shift_to(X); addmul(*T,iM); } }
  };
}
////////////////////////////////////////////////////////////////////////////////
#undef FI
#undef op

#endif // included_tayl_h
