
#include <iostream>
#include <iomanip>
#include <strstream>
#include <math.h>
#include <body.h>
#include <falcON.h>
#include <time.h>

using namespace std;
using namespace fasttree;

const double Pi         = 3.14159265358979323846264338328;
const double Pih        = 0.5 * Pi;
const double TPi        = 2.0 * Pi;

double ms_Force_gamma(const double gamma)
{
  register double x=5-3*gamma,f;
  if(x < 0.) return 0.;
  f =1/x;
  f-=4/(x+=1);
  f+=6/(x+=1);
  f-=4/(x+=1);
  f+=1/(x+=1);
  return (3-gamma)*f;
}

double ms_density_gamma(const double gamma)
{
  register double x=3-3*gamma,f;
  if(x < 0.) return 0.;
  f = 1/x;
  f-= 8/(x+=1);
  f+=28/(x+=1);
  f-=56/(x+=1);
  f+=70/(x+=1);
  f-=56/(x+=1);
  f+=28/(x+=1);
  f-= 8/(x+=1);
  f+= 1/(x+=1);
  return cube(3-gamma)*f/square(4*Pi);
}

int main(int argc, char* argv[])
{
  if(argc != 9 && argc != 10 && argc != 11 && argc != 12 && argc != 13) {
    cerr<<
      " \"TestFalcON MOD gamma N S THE C K EPS [Nc] [T] [Nt] [Nf]\" with \n"
#ifdef TWODIMENSIONAL
      " MOD = 3/4      : Kuzmin disk / homogeneous disk\n"
#else
      " MOD = 0/1/2/3/4: hom. sphere / Plummer / gamma-model"
      " / Kuzmin disk / hom. disk\n"
#endif
      " gamma               : parameter of model (if any)\n"
      " N                   : number of positions\n"
      " S = long            : seed for RNG\n"
      " THE                 : opening angle, if THE<0, use theta=const\n"
      " C = 0/1             : center root on origin or not ?\n"
      " K = n/(10+n)/(20+n) : F_n / K_n / P_n softening kernel\n"
      " EPS                 : max / fixed softening length\n"
      " Nc (default 1)      : don't split cells with N <= Nc bodies\n"
      " T  (default 0) 0/1  : dump nodes to tree.cells and tree.souls\n"
      " Nt (default 0)      : the Nt last bodies are test-particles\n"
      " Nf (default 0)      : use only Nf sinks\n";
    exit(1);
  }

  register clock_t  cpu0 = clock(), cpu1;
  register real     time_setup, time_grow, time_esti;
  const real        rmax = 1.e3;
  int               p=1, MOD, C, KER, T=0;
  kern_type         K;
  soft_type         Soft;
  long              Seed;
  unsigned          N, NS, NT=0, NNT, Ncrit=1;
  real              EPS, TH, GAM;
  istream          *in;

  { in = new istrstream(argv[p],strlen(argv[p])); *in>>MOD;     delete in; p++; }
  { in = new istrstream(argv[p],strlen(argv[p])); *in>>GAM;     delete in; p++; }
  { in = new istrstream(argv[p],strlen(argv[p])); *in>>N;       delete in; p++; }
  { in = new istrstream(argv[p],strlen(argv[p])); *in>>Seed;    delete in; p++; }
  { in = new istrstream(argv[p],strlen(argv[p])); *in>>TH;      delete in; p++; }
  { in = new istrstream(argv[p],strlen(argv[p])); *in>>C;       delete in; p++; }
  { in = new istrstream(argv[p],strlen(argv[p])); *in>>KER;     delete in; p++; }
  { in = new istrstream(argv[p],strlen(argv[p])); *in>>EPS;     delete in; p++; }
  if(argc>9)
    { in = new istrstream(argv[p],strlen(argv[p])); *in>>Ncrit; delete in; p++; }
  if(argc>10)
    { in = new istrstream(argv[p],strlen(argv[p])); *in>>T;     delete in; p++; }
  if(argc>11)
    { in = new istrstream(argv[p],strlen(argv[p])); *in>>NT;    delete in; p++; }
  if(NT >= N) { cerr<<" Nt >= N\n"; exit(1); }
  NNT = N-NT;
  if(argc>12)
    { in = new istrstream(argv[p],strlen(argv[p])); *in>>NS;    delete in; p++; }
  else NS = N;

  bodies *BODIES = new bodies(N);
  vect  *A = new vect[N];
  real *RH = new real[N];
  srand48(Seed);
  switch(KER) {
  case 20: K = fasttree::p0; break;
  case 21: K = fasttree::p1; break;
  case 22: K = fasttree::p2; break;
  case 23: K = fasttree::p3; break;
  case 10: K = fasttree::k0; break;
  case 11: K = fasttree::k1; break;
  case 12: K = fasttree::k2; break;
  case 13: K = fasttree::k3; break;
  case  0: K = fasttree::f0; break;
  case  1: K = fasttree::f1; break;
  case  2: K = fasttree::f2; break;
  case  3: K = fasttree::f3; break;
  default: K = fasttree::newton; break;
  }
  Soft = fasttree::global;
  EPS  = abs(EPS);

  const    real g1=1.-GAM, g2=2.-GAM, ig2=1./g2, g3=3-GAM, ig3=1./g3;
  register real tm,r,R,phi,fr,P,rh;
#ifndef TWODIMENSIONAL
  register real cth;
#endif
  register double mi = one/real(NNT);
  vect origin=0., *porigin = C? &origin : 0;
  LoopBodies(bodies,BODIES,Bi) {
    do {
      switch(MOD) {
#ifndef TWODIMENSIONAL
      case 0:  // homogeneous
	r  = pow(drand48(),1./3.);
	P  = 0.5*(3*r*r-1);
	fr =-1.;
	rh = 0.75/Pi;
	break;
      case 1:  // Plummer
	tm = pow(drand48(),2./3.);
	tm/= 1.-tm;                    // r^2
	r  = sqrt(tm);
	P  =-1./sqrt(tm+1);
	fr = P*P*P;
	rh = 0.75/Pi*pow_int(-P,5);
	break;
      case 2:  // gamma-model
	tm = pow(drand48(),ig3);
	r  = tm / (1.-tm);
	P  = (GAM==2.)?  log(tm) : -ig2*(1-pow(tm,g2));
	fr =-pow(tm,g1) * square(1-tm) / r;
	rh = 0.25*g3/Pi * pow_int(1.-tm,4)/POW(tm,GAM);
	break;
#endif
      case 3:
	tm = drand48();
	r  = sqrt((2-tm)*tm)/(1.-tm);
	P  =-1./sqrt(r*r+1);
	fr = P*P*P;
	rh =-0.5/Pi*fr;
	break;
      case 4:
	tm = drand48();
	r  = sqrt(tm);
	P  = 2*(r*r/3.-2);
	fr =-4*r/3.;
	rh = 1.0/Pi;
	break;
       default:
	cerr<<" unknown model\n";
	exit(1);
      }
    } while(r>rmax);
#ifdef TWODIMENSIONAL
    R = r;
#else
    cth = (MOD==3) ? zero : 2*drand48()-1;
    R   = r*sqrt(1.-cth*cth);
    Bi.pos(2) = r * cth;
#endif
    phi        = TPi*drand48();
    Bi.pos(0) = R * sin(phi);
    Bi.pos(1) = R * cos(phi);
    if(index(Bi)<NNT) Bi.mass() = mi;
    else      Bi.mass() = zero;
    if(index(Bi)<NS) Bi.flag_as_sink();
    else     Bi.unflag_sink();
    A [index(Bi)] = fr * pos(Bi);   // exact force
    RH[index(Bi)] = rh;             // exact density
  }
//   // create file with masses & positions
//   ofstream out("positions.dat");
//   out<<N<<"\n";
//   for(Bi=B,i=0; i!=N; ++Bi,++i)
//     out<<setprecision(12)<<mi<<" "<<pos(Bi)<<"\n";
//   out.close();
//   //                                    
  cpu1 = clock();
  time_setup = (cpu1 - cpu0) / real(CLOCKS_PER_SEC);
  cpu0 = cpu1;
  cout<<"\n"<<setprecision(6)
      <<" time needed for set up of X_i:           "<<time_setup<<"\n";

  falcON falcon(BODIES,EPS,TH,K);
  falcon.grow(Ncrit,porigin);
  cpu1 = clock();
  time_grow = (cpu1 - cpu0) / real(CLOCKS_PER_SEC);
  cpu0 = cpu1;
  cout<<setprecision(6)
      <<" time needed for falcON::grow():          "<<time_grow<<"\n";
  if(T) falcon.dump_nodes("tree.cells","tree.souls");

  cpu0 = clock();
  falcon.approximate_gravity();
  cpu1 = clock();
  time_esti = (cpu1 - cpu0) / real(CLOCKS_PER_SEC);
  cout<<setprecision(6)
      <<" time needed for falcON::approximate():   "<<time_esti<<"\n";
  falcon.stats(cout);
  cout<<"\n";

  register real ase=0.,damx=0.,dacc;
  register vect SF=0.;
  LoopBodies(bodies,BODIES,Bi) {
    dacc = norm(A[index(Bi)] - acc(Bi));
    ase += dacc;
    damx = fasttree::max(damx,dacc);
    SF  += mass(Bi)*acc(Bi);
//     cerr<<setfill(' ')
// 	<<setw(13)<<setprecision(8)<<Bi.pos()[0]<<" "
// 	<<setw(13)<<setprecision(8)<<Bi.pos()[1]<<" "
// 	<<setw(13)<<setprecision(8)<<Bi.pos()[2]<<" "
// 	<<setw(13)<<setprecision(8)<<Bi.acc()[0]<<" "
// 	<<setw(13)<<setprecision(8)<<Bi.acc()[1]<<" "
// 	<<setw(13)<<setprecision(8)<<Bi.acc()[2]<<endl;
  }
  ase /= double(N);
  switch(MOD){
  case 1: ase *= 13.125; break;
  case 3: ase *= 17.5;   break;
  case 4: ase *= 1.125;  break;
  case 0: ase /= 0.6;    break;
  case 2: if(GAM<1.66) ase /= ms_Force_gamma(GAM); break;          
  default: break;
  }
  if(MOD==2 && GAM>=1.66)
    cout<<setprecision(10)<<" ASE(F)           = "<< ase <<"\n";
  else
    cout<<setprecision(10)<<" ASE(F)/<F^2>     = "<< ase <<"\n";
  cout<<" max (dF)^2       = "<< damx<<"\n"
      <<" Sum m_i F_i      = "<< SF  <<"\n";
//  cout<<endl;
//   for(;;) {
//     cerr<<"i = "; cin>>i;
//     cerr<<" "<<acc(B+i)<<" "<<neighbours(B+i)<<"\n";
//   }
  delete BODIES;
}
