//-----------------------------------------------------------------------------+
//                                                                             |
// falcON.cc                                                                   |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines all function members of class tree                                  |
//                                                                             |
// Contents                                                                    |
//                                                                             |
// 1  Constructor, destructor and such                                         |
// 2  tree growth and re-use                                                   |
// 3  information and statistics                                               |
// 4  estimation of density and such                                           |
// 5  approximation of gravity                                                 |
// 6  neighbour finding and counting                                           |
//                                                                             |
//-----------------------------------------------------------------------------+
#include <falcON.h>                        // class falcON                      
#include <body.h>                          // class body and all its members    
#include <tree/soul.h>                     // class soul and all its members    
#include <tree/cell.h>                     // class cell and all its members    
#include <tree/tree.h>                     // class tree                        
#include <tree/stic.h>                     // support for SPH & sticky particles
#include <tree/tayl.h>                     // Taylor series & stuff             
#include <tree/open.h>                     // class opening                     
#include <tree/grav.h>                     // class interactor                  
#include <tree/iact.h>                     // generic interaction algorithm     
#include <tree/stat.h>                     // statistics & error messages       
#include <iomanip.h>                       // C++ I/O formatting etc            
using std::ostream;                        // we'll use ostream                 
using std::ofstream;                       // and ofstream                      
namespace fasttree {
  typedef tree<cell> tree_type;            // the tree type used 
}
using namespace fasttree;                  // make all tree stuff available     
//=============================================================================+
//                                                                             |
// 1 CONSTRUCTORS, DESTRUCTOR AND SUCH                                         |
// ===================================                                         |
//                                                                             |
//=============================================================================+
falcON::falcON(const bodies*   b,
	       const real      e,
	       const real      th,
	       const kern_type t,
	       const uint      d) :
  STATE(unset),
  BODIES(b), TREE(0),
  Nb(BODIES->N_body()), Ncrit(0),
  FL(0), M(0), EP(0), X(0), PH(0), RH(0), A(0),
  OPENING(new opening(abs(th),(NDIM==2 || th<zero))),
  STATS  (new TreeStats()),
  DMAX   (d)
{
  reset_softening(e,t);
}
//------------------------------------------------------------------------------
falcON::falcON(const int      *f,
	       const areal    *m,
	       const areal    *x[NDIM],
	       const areal    *ep,
	             areal    *a[NDIM],
	             areal    *ph,
	             areal    *rh,
	       const unsigned  n,
	       const real      e,
	       const real      th,
	       const kern_type t,
	       const unsigned  d) :
  STATE(unset),
  BODIES(0), TREE(0),
  Nb(n), Ncrit(0),
  FL(f),M(m),EP(ep),X(x),PH(ph),RH(rh),A(a),
  OPENING(new opening(abs(th),(NDIM==2 || th<zero))),
  STATS(new TreeStats()),
  DMAX(d)
{
  reset_softening(e,t);
}
//------------------------------------------------------------------------------
falcON::~falcON()
{
  if(TREE) delete TREE;
  delete OPENING;
  delete STATS;
}
//------------------------------------------------------------------------------
void falcON::reset_softening (const real e, const kern_type t) const
{
  SOFTENING = (e<zero)? individual : global;
  EPS       = (e<zero)? zero : e;
  KERNEL    = t;
}
//------------------------------------------------------------------------------
void falcON::reset_opening(const real t) const
{
  OPENING->reset(abs(t),t<zero);
}
//------------------------------------------------------------------------------
inline cell* falcON::root()
{
  return TREE->root();
}
//------------------------------------------------------------------------------
inline const cell* falcON::root() const
{
  return TREE->root();
}
//=============================================================================+
//                                                                             |
// 2 TREE GROWTH AND RE-USE                                                    |
// ========================                                                    |
//                                                                             |
//=============================================================================+
inline real bmax(const cell*C)
  // This routines returns the distance from the cell's cofm to its most        
  // distant corner.                                                            
{
  register real bq=zero;                           // initialize return variable
  for(register indx i=0; i<NDIM; i++)              // loop dimensions           
    bq += square(radius(C)+abs(cofm(C)[i]-center(C)[i])); // add in square      
  return SQRT(bq);                                 // return sqrt of total      
}
//------------------------------------------------------------------------------
inline void eval_basic_source(const bool reuse, tree_type*T)
  // This routines computes for all cells of tree T:                            
  //  - mass, 1/mass, cofm, rmax (needed for almost everything)                 
  // To be uses AFTER souls have been updated                                   
{
  register real  mon;                              // monopole                  
  register vect  dip;                              // dipole                    
  LoopCellsUp(tree_type,T,Ci) {                    // loop cells upwards >      
    mon = zero;                                    //   reset monopole          
    dip = zero;                                    //   reset dipole            
    LoopCellKids(cell,Ci,c) {                      //   loop sub-cells c >      
	mon += mass(c);                            //       sum up monopole     
	dip += cofm(c);                            //       sum up dipole       
    }                                              //   <                       
    LoopSoulKids(cell,Ci,s) {                      //   loop sub-souls s >      
	mon += mass(s);                            //       sum up monopole     
	dip += mass(s)*cofm(s);                    //       sum up dipole       
    }                                              //   <                       
    Ci->mass() = mon;                              //   set mass                
    Ci->imass()= (mon==zero)? zero:one/mon;        //   set 1/mass              
    Ci->cofm() = dip;                              //   set dipole              
  }                                                // <                         
  register real dmax, x, Xq;                       // d_max, auxiliary vars     
  LoopCellsUp(tree_type,T,Ci) {                    // loop cells upwards >      
    Ci->cofm() *= imass(Ci);                       //   cofm = dipole/mass      
    dmax = zero;                                   //   reset dmax              
    LoopSoulKids(cell,Ci,s)                        //   loop sub-souls s >      
      update_max(dmax,dist_sq(cofm(s),cofm(Ci)));  //     update d_max^2 <      
    if(has_soul_children(Ci)) dmax = SQRT(dmax);   //   d_max due to sub-souls  
    LoopCellKids(cell,Ci,c) {                      //   loop sub-cells c >      
      Xq = dist_sq(cofm(c),cofm(Ci));              //     distance^2            
      x  = dmax - rmax(c);                         //     d=distance + r_i      
      if(zero > x || Xq>x*x)                       //     IF(d>d_max)      >    
	dmax = SQRT(Xq) + rmax(c);                 //       set d_max = d  <    
    }                                              //   <                       
    Ci->rmax() = reuse? dmax:min(dmax,bmax(Ci));   //   assign r_max            
  }                                                // <                         
}
//------------------------------------------------------------------------------
void falcON::reuse()
{
  // This routine uses the existing tree and                                    
  // -  re-sets the souls' positions, masses, and source flags                  
  // -  re-evaluates the basic source properties of the cells                   
  // and pre-compute basic source properties of cells                           
  if(TREE == 0) {
    warning(" cannot reuse(), no old tree existent");
    return grow();
  }
  if(BODIES) LoopSouls(tree_type,TREE,Si) Si->set_up(BODIES);// re-set souls    
  else       LoopSouls(tree_type,TREE,Si) Si->set_up(M,X);   // re-set souls    
  eval_basic_source(true,TREE);                    // re-evaluate cells' props  
  STATE=built;                                     // set tree state            
}
//------------------------------------------------------------------------------
void falcON::grow(const uint n, const vect*pc)
  // This routines grows a fresh tree & evaluates the cells' basic source       
  // properties.                                                                
{
  Ncrit = max(n,1u);
  if(TREE) delete TREE;
  if(BODIES)
    TREE = new tree_type(BODIES,pc,Ncrit,DMAX);
  else {
    TREE = new tree_type(FL,X,Nb,pc,Ncrit,DMAX);
    LoopSouls(tree_type,TREE,Si) Si->set_mass(M); 
  }
  eval_basic_source(false,TREE);                   // evaluate cells' props     
  STATE=built;                                     // set tree state            
}
//=============================================================================+
//                                                                             |
// 3 INFORMATION AND STATISTICS                                                |
// ============================                                                |
//                                                                             |
//=============================================================================+
const char* falcON::opening_method() const {
  return OPENING->method();
}
//------------------------------------------------------------------------------
real falcON::theta(const real m) const {
  return OPENING->theta(m);
}
//------------------------------------------------------------------------------
const fasttree::uint &falcON::No_bodies() const {
  return BODIES->N_body();
}
//------------------------------------------------------------------------------
fasttree::uint falcON::No_bodies_used() const {
  return TREE->N_souls();
}
//------------------------------------------------------------------------------
fasttree::uint falcON::No_cells_used() const {
  return TREE->N_cells();
}
//------------------------------------------------------------------------------
fasttree::uint falcON::No_zombie_cells() const {
  return TREE->N_zombies();
}
//------------------------------------------------------------------------------
fasttree::uint falcON::No_zombie_bodies() const {
  register uint  n=0;
  if(BODIES) {
    LoopSouls(tree_type,TREE,Si) if(!is_in_tree(BODIES->flag(mybody(Si)))) n++;
  } else {
    LoopSouls(tree_type,TREE,Si) if(!is_in_tree(FL[mybody(Si)])) n++;
  }
  return n;
}
//------------------------------------------------------------------------------
namespace fasttree {
  inline void dump_nodes(const cell*Ci, ostream*oc, ostream*os,
			 const cell*C0, const soul*S0)
    // RECURSIVE
  {
    if(oc) { Ci->dump(*oc,S0,C0); *oc<<"\n"; }
    LoopCellKids(cell,Ci,c) dump_nodes(c,oc,os,C0,S0);
    if(os) LoopSoulKids(cell,Ci,s) { s->dump(*os,S0); *os<<"\n"; }
  }
  //----------------------------------------------------------------------------
  inline int trick(const real x, const int w)
  {
    if(x<0.001)  return max(1,w-5);
    if(x<0.01)   return max(1,w-4);
    if(x<0.1)    return max(1,w-3);
    if(x<one)    return max(1,w-2);
    if(x<ten)    return max(1,w-1);
    return w;
  }
}
//------------------------------------------------------------------------------
void falcON::dump_nodes(const char* fcells, const char* fleafs) const
{
  ofstream *oc = fcells? new ofstream(fcells) : 0;
  ofstream *os = fleafs? new ofstream(fleafs) : 0;
  if(oc) { cell::dump_head(*oc); (*oc)<<"\n"; }
  if(os) { soul::dump_head(*os); (*os)<<"\n"; }
  if(oc || os) ::dump_nodes(root(),oc,os,
			    TREE->begin_cells(), TREE->begin_souls());
  if(oc) { oc->flush(); delete oc; }
  if(os) { os->flush(); delete os; }
}
//------------------------------------------------------------------------------
const char* falcON::describe(const kern_type KER)
{
  switch(KER) {
  case k0: return "K0";
  case k1: return "K1";
  case k2: return "K2";
  case k3: return "K3";
  case f0: return "F0";
  case f1: return "F1";
  case f2: return "F2";
  case f3: return "F3";
  case p0: return "P0==Plummer";
  case p1: return "P1";
  case p2: return "P2";
  case p3: return "P3";
  case newton:
  default: return "Newtonian";
  }
}
//------------------------------------------------------------------------------
void falcON::stats(ostream& out) const
{
  out<<"\n state:                ";
  if(STATE&unset)  out<<" unset";
  if(STATE&built)  out<<" built";
  if(STATE&reused) out<<" reused";
  out<<"\n";
  if(STATE!=unset) {
    out<<" root center:           "<<center(root())       <<"\n"
       <<" root radius:           "<<radius(root())       <<"\n"
       <<" bodies loaded:         "<<number(root())       <<"\n"
       <<" total mass:            "<<mass  (root())       <<"\n"
       <<" N_crit:                "<<Ncrit                <<"\n"
       <<" cells used:            "<<TREE->N_cells()      <<"\n"
       <<" zombies elimitated:    "<<TREE->N_zombies()    <<"\n"
       <<" maximum depth:         "<<TREE->depth()        <<"\n"
       <<" current theta :        "<<OPENING->theta_min() <<"\n"
       <<" current opening :      "<<OPENING->method()    <<"\n";
    if(SOFTENING == global) 
      out<<" softening:             global\n"
	 <<" softening length:      "<< EPS               <<"\n";
    else
      out<<" softening:             individual\n";
    out<<" softening kernel:      "<<describe_kernel()    <<"\n";
  }
  register uint total = STATS->BB_iacts()
                      + STATS->CB_taylor_iacts()
                      + STATS->CB_direct_iacts()
                      + STATS->CC_taylor_iacts()
                      + STATS->CC_direct_iacts()
                      + STATS->CS_direct_iacts(), part;
  register real percent;
  out<<" interaction statitics:\n"
     <<"     type          approx   direct      total\n"
     <<" # body-body :          - ";
  part    = STATS->BB_iacts();
  percent = 100.*part/real(total);
  out<<setw( 8)<<STATS->BB_iacts()<<" "
     <<setw(10)<<part<<" = "
     <<setprecision(trick(percent,5))<<setw(8)<<percent<<"%\n"
     <<" # cell-body : ";
  part    = STATS->CB_taylor_iacts()
	   +STATS->CB_direct_iacts();
  percent = 100.*part/real(total);
  out<<setw(10)<<STATS->CB_taylor_iacts()<<" "
     <<setw( 8)<<STATS->CB_direct_iacts()<<" "
     <<setw(10)<<part<<" = "
     <<setprecision(trick(percent,5))<<setw(8)<<percent<<"%\n"
     <<" # cell-cell : ";
  part    = STATS->CC_taylor_iacts()
	   +STATS->CC_direct_iacts();
  percent = 100.*part/real(total);
  out<<setw(10)<<STATS->CC_taylor_iacts()<<" "
     <<setw( 8)<<STATS->CC_direct_iacts()<<" "
     <<setw(10)<<part<<" = "
     <<setprecision(trick(percent,5))<<setw(8)<<percent<<"%\n"
     <<" # cell-self :          - ";
  part    = STATS->CS_direct_iacts();
  percent = 100.*part/real(total);
  out<<setw( 8)<<STATS->CS_direct_iacts()<<" "
     <<setw(10)<<part<<" = "
     <<setprecision(trick(percent,5))<<setw(8)<<percent<<"%\n"
     <<" # total     : ";
  part=STATS->BB_iacts()
      +STATS->CB_direct_iacts()
      +STATS->CC_direct_iacts()
      +STATS->CS_direct_iacts();
  out<<setw(10)<<STATS->CB_taylor_iacts()+STATS->CC_taylor_iacts()<<" "
     <<setw( 8)<<part<<" "
     <<setw(10)<<total<<" =  100.000%\n";
}
//------------------------------------------------------------------------------
fasttree::uint falcON::BB_interactions() const {
  return STATS->BB_iacts();
}
//------------------------------------------------------------------------------
fasttree::uint falcON::MB_interactions() const {
  return STATS->CB_direct_iacts() +STATS->CC_direct_iacts()
        +STATS->CS_direct_iacts();
}
//------------------------------------------------------------------------------
fasttree::uint falcON::CB_interactions() const {
  return STATS->CB_taylor_iacts();
}
//------------------------------------------------------------------------------
fasttree::uint falcON::CC_interactions() const {
  return STATS->CC_taylor_iacts();
}
//------------------------------------------------------------------------------
fasttree::uint falcON::total_interactions () const {
  return BB_interactions()+MB_interactions()
        +CB_interactions()+CC_interactions();
}
//------------------------------------------------------------------------------
const fasttree::vect& falcON::root_center() const { return center(root()); }
const fasttree::real& falcON::root_radius() const { return radius(root()); }
const fasttree::uint& falcON::root_depth () const { return TREE->depth(); }
const fasttree::uint& falcON::root_number() const { return number(root()); }
const fasttree::real& falcON::root_mass  () const { return mass  (root()); }
//-----------------------------------------------------------------------------+
//                                                                             |
// 4 ESTIMATION OF DENSITY AND SUCH                                            |
// ================================                                            |
//                                                                             |
//-----------------------------------------------------------------------------+
namespace fasttree {
  inline void pass_flags_up(tree_type*T, const bodies*B)
  {
    LoopSouls  (tree_type,T,Si) Si->copy_sink_flag(B);
    LoopCellsUp(tree_type,T,Ci) Ci->update_sink_flag();
  }
  //----------------------------------------------------------------------------
  inline void pass_flags_up(tree_type*T, const int *FL)
  {
    LoopSouls  (tree_type,T,Si) Si->copy_sink_flag(FL);
    LoopCellsUp(tree_type,T,Ci) Ci->update_sink_flag();
  }
  //----------------------------------------------------------------------------
  inline void guess_N(cell*Ci, real n, const uint& Nx)
  {
#if NDIM==3
    if(number(Ci)>Nx || n==zero) n = number(Ci) / (NSUB*cube(radius(Ci)));
#else  // NDIM==3
    if(number(Ci)>Nx || n==zero) n = number(Ci) / (NSUB*square(radius(Ci)));
#endif // NDIM==3
    LoopCellKids(cell,Ci,c) if(is_sink(c)) guess_N(c,n,Nx);
    LoopSoulKids(cell,Ci,s) if(is_sink(s)) s->rho() = n;
  }
  //----------------------------------------------------------------------------
  inline void guess_sd(cell*Ci, real sd, const uint& Nx)
  {
    if(number(Ci)>Nx || sd==zero) sd = mass(Ci) / (4*square(radius(Ci)));
    LoopCellKids(cell,Ci,c) if(is_sink(c)) guess_sd(c,sd,Nx);
    LoopSoulKids(cell,Ci,s) if(is_sink(s)) s->rho() = sd;
  }
  //----------------------------------------------------------------------------
  inline void guess_rho(cell*Ci, real rh, const uint& Nx)
  {
#if NDIM==3
    if(number(Ci)>Nx || rh==zero) rh = mass(Ci) / (NSUB*cube(radius(Ci)));
#else  // NDIM==3
    if(number(Ci)>Nx || rh==zero) rh = mass(Ci) / (NSUB*square(radius(Ci)));
#endif // NDIM==3
    LoopCellKids(cell,Ci,c) if(is_sink(c)) guess_rho(c,rh,Nx);
    LoopSoulKids(cell,Ci,s) if(is_sink(s)) s->rho() = rh;
  }
}
//------------------------------------------------------------------------------
void falcON::estimate_n(const uint Nx)
{
  if(BODIES) pass_flags_up(TREE,BODIES);
  else       pass_flags_up(TREE,FL);               // update flags              
  guess_N(root(),zero,Nx);                         // guess n_i                 
  if(BODIES == 0)
    LoopSouls(tree_type,TREE,Si)
      if(is_sink(Si)) Si->update_dens(RH);         // copy n_i into array RH    
}
//------------------------------------------------------------------------------
void falcON::estimate_sd(const uint Nx)
{
  if(BODIES) pass_flags_up(TREE,BODIES);
  else       pass_flags_up(TREE,FL);               // update flags              
  guess_sd(root(),zero,Nx);                        // guess SD_i                
  if(BODIES == 0)
    LoopSouls(tree_type,TREE,Si)
      if(is_sink(Si)) Si->update_dens(RH);         // copy n_i into array RH    
}
//------------------------------------------------------------------------------
void falcON::estimate_rho(const uint Nx)
{
  if(BODIES) pass_flags_up(TREE,BODIES);
  else       pass_flags_up(TREE,FL);               // update flags              
  guess_rho(root(),zero,Nx);                       // guess n_i                 
  if(BODIES == 0)
    LoopSouls(tree_type,TREE,Si)
      if(is_sink(Si)) Si->update_dens(RH);         // copy n_i into array RH    
}
//-----------------------------------------------------------------------------+
//                                                                             |
// 5 APPROXIMATION OF GRAVITY                                                  |
// ==========================                                                  |
//                                                                             |
//-----------------------------------------------------------------------------+
namespace fasttree {
  inline uint pass_flags_and_eps_up(const opening   *OPENING,
				    const soft_type &SOFTENING,
				          tree_type *T)
    // preparation: passing flags & eps up the tree, set R_crit                 
  {
    register uint n=0u;
    if(SOFTENING == individual) {
      register real Emin, Emax;
#ifdef SOFT_MEAN
      register real Eps;
#endif
      LoopCellsUp(tree_type,T,Ci) {
	Ci->set_rcrit(OPENING->invtheta(mass(Ci)));
	Ci->reset_sink_flag();
	Emin = zero;
	Emax = zero;
#ifdef SOFT_MEAN
	Eps  = zero;
#endif
	LoopCellKids(cell,Ci,c) {
	  Ci->add_sink_flag(c);
	  if(Emin==zero || emin(c)<Emin) Emin = emin(c);
	  if(Emax==zero || emax(c)>Emax) Emax = emax(c);
#ifdef SOFT_MEAN
	  Eps += eps(c);
#endif
	}
	LoopSoulKids(cell,Ci,s) {
	  Ci->add_sink_flag(s);
	  if(Emin==zero || eps(s)<Emin) Emin = eps(s);
	  if(Emax==zero || eps(s)>Emax) Emax = eps(s);
#ifdef SOFT_MEAN
	  Eps += mass(s)*eps(s);
#endif
	}
	if(is_sink(Ci)) n++;
	Ci->emin() = Emin;
	Ci->emax() = Emax;
#ifdef SOFT_MEAN
	Ci->eps()  = Eps;
#else
	Ci->eps()  = Emin;
#endif
      }
#ifdef SOFT_MEAN
      LoopCellsUp(tree_type,T,Ci) Ci->eps() *= imass(Ci);
#endif
    } else { // SOFTENING == fixed, no cell-intern epsila
      LoopCellsUp(tree_type,T,Ci) {
	Ci->set_rcrit(OPENING->invtheta(mass(Ci)));
	Ci->update_sink_flag();
	if(is_sink(Ci)) n++;
      }
    }
    return n;
  }
  //----------------------------------------------------------------------------
  inline uint pass_flags_and_eps_up(const opening   *OPENING,
				    const soft_type &SOFTENING,
				          tree_type *T,
				    const bodies    *B) {
    LoopSouls(tree_type,T,Si) Si->reset_grav(B);
    return pass_flags_and_eps_up(OPENING,SOFTENING,T);
  }
  //----------------------------------------------------------------------------
  inline uint pass_flags_and_eps_up(const opening   *OPENING,
				    const soft_type &SOFTENING,
				          tree_type *T,
				    const int       *FL,
				    const areal     *EP=0) {
    if(EP) LoopSouls(tree_type,T,Si) Si->reset_grav(FL,EP);
    else   LoopSouls(tree_type,T,Si) Si->reset_grav(FL);
    return pass_flags_and_eps_up(OPENING,SOFTENING,T);
  }
  //----------------------------------------------------------------------------
  inline void evaluate_quad(tree_type*T)
    // pre-computes quadrupole moments for all cells (needed for gravity only)  
  {
    register vect Xi;
    register matr quad,Qi;
    LoopCellsUp(tree_type,T,Ci) {                  // loop tree upwards >       
      quad = zero;                                 //   reset quadrupole        
      LoopSoulKids(cell,Ci,s) {                    //   loop over subsouls >    
	Xi    = cofm(s)-cofm(Ci);                  //     vector to subsoul     
	Qi.outprod(Xi);                            //     add quad due to       
	Qi   *= mass(s);                           //      ub-node's            
	quad += Qi;                                //     position              
      }                                            //   <                       
      LoopCellKids(cell,Ci,c) {                    //   loop over subcells >    
	Xi    = cofm(c)-cofm(Ci);                  //     vector to subcell     
	quad += quadru(c) ;                        //     add sub-qu            
	Qi.outprod(Xi);                            //     add quad due to       
	Qi   *= mass(c);                           //     sub-node's            
	quad += Qi;                                //     position              
      }                                            //   <                       
      Ci->quadru() = quad;                         //   assign quadrupole       
    }                                              // <                         
    LoopCellsDown(tree_type,T,Ci)                  // loop cells                
      Ci->quadru() *= imass(Ci);                   //   normalize quadrupole <  
  }
  //----------------------------------------------------------------------------
  inline void evaluate_grav(cell       *Ci,
			    TaylorGrav  G)
    // evaluation algorithm for gravity                                         
  {
    G.shift_and_add(cofm(Ci),coeffs(Ci),imass(Ci));
    LoopSoulKids(cell,Ci,s) if(is_sink(s)) G.extract_grav(s);
    LoopCellKids(cell,Ci,c) if(is_sink(c)) evaluate_grav(c,G);
  }
}
//------------------------------------------------------------------------------
// approximation: prepare, interact, and collect
void falcON::approximate_gravity()
{
  register uint Nu = BODIES?
    pass_flags_and_eps_up(OPENING,SOFTENING,TREE,BODIES) :
    pass_flags_and_eps_up(OPENING,SOFTENING,TREE,FL,EP);
  if(Nu==0) {
    warning(" falcON::approximate_gravity(): no body wants updating");
    return;
  }
  T0 = new TaylorCoeffs[Nu];
  register TaylorCoeffs *Ti=T0;
  LoopCellsDown(tree_type,TREE,Ci) if(is_sink(Ci)) Ci->coeffs() = Ti++;
  STATS->reset();
  gravity_kernel K(STATS,KERNEL,EPS,SOFTENING);
  evaluate_quad(TREE);
  MutualInteract<gravity_kernel,cell,soul>(root(),TREE->depth(),&K);
  TaylorGrav G(cofm(root()));
  LoopSouls(tree_type,TREE,Si) if(is_sink(Si)) Si->normalize_grav();
  evaluate_grav(root(),G);
  delete[] T0;
  T0 = 0;
  if(BODIES == 0) LoopSouls(tree_type,TREE,Si)
		    if(is_sink(Si)) Si->update_grav(A,PH);
}
//------------------------------------------------------------------------------
// exact gravity by direct summation
void falcON::exact_gravity()
{
  register uint Nu = pass_flags_and_eps_up(OPENING,SOFTENING,TREE,FL,EP);
  if(Nu==0) {
    warning(" tree::exact_gravity(): no body wants updating");
    return;
  }
  STATS->reset();
  gravity_kernel K(STATS,KERNEL,EPS,SOFTENING);
  K.direct_summation(root());
  if(BODIES) {
    LoopSouls(tree_type,TREE,Si) if(is_sink(Si))
      Si->normalize_grav();
  } else {
    LoopSouls(tree_type,TREE,Si) if(is_sink(Si)) {
      Si->normalize_grav();
      Si->update_grav(A,PH);
    }
  }
}
//-----------------------------------------------------------------------------+
//                                                                             |
// 6 NEIGHBOUR FINDING AND COUNTING                                            |
// ================================                                            |
//                                                                             |
//-----------------------------------------------------------------------------+
void falcON::make_iaction_list(      elem_pair*bl,
			       const uint      nl,
			             uint     &na,
			       const real      tau)
{
  if(BODIES == 0) error(" mismatch in using bodies/arrays");
  if(tau < zero) {             // SPH
    LoopSouls(tree_type,TREE,Si) if(is_sph(BODIES->flag(mybody(Si)))) {
      flag_for_subtree(Si);
      Si->add_body_flag(BODIES);
    }
    sticky_tree      stree(TREE);
    neighbour_finder sfind(nl,bl);
    stree.prepare_sph();
    MutualInteract<neighbour_finder,sticky_cell,sticky_soul>
      (stree.root(),stree.depth(),&sfind);
    na = sfind.actual_size_of_list();
  } else {                     // sticky particles
    LoopSouls(tree_type,TREE,Si) if(is_sticky(BODIES->flag(mybody(Si)))) {
      flag_for_subtree(Si);
      Si->add_body_flag(BODIES);
    }
    sticky_tree   stree(TREE);
    sticky_finder sfind(tau,nl,bl);
    stree.prepare_sticky();
    MutualInteract<sticky_finder,sticky_cell,sticky_soul>
      (stree.root(),stree.depth(),&sfind);
    na = sfind.actual_size_of_list();
  }
}
//-----------------------------------------------------------------------------+
void falcON::count_sph_neighbours()
{
  if(BODIES == 0) error(" mismatch in using bodies/arrays");
  LoopSouls(tree_type,TREE,Si) if(is_sph(BODIES->flag(mybody(Si)))) {
    flag_for_subtree(Si);
    Si->add_body_flag(BODIES);
  }
  sticky_tree       stree (TREE);
  neighbour_counter scount;
  stree.prepare_sph();
  MutualInteract<neighbour_counter,sticky_cell,sticky_soul>
    (stree.root(),stree.depth(),&scount);
  stree.set_num(TREE->begin_souls());
}
//-----------------------------------------------------------------------------+
void falcON::make_iaction_list(      elem_pair*bl,
			       const uint      nl,
			             uint     &na,
			       const areal    *s,
			       const areal    *v[NDIM],
			       const real      tau)
{
  if(BODIES) error(" mismatch in using bodies/arrays");
  if(tau < zero) {             // SPH
    LoopSouls(tree_type,TREE,Si) if(is_sph(FL[mybody(Si)])) {
      flag_for_subtree(Si);
      Si->add_body_flag(FL);
    }
    sticky_tree      stree(TREE,s,v);
    neighbour_finder sfind(nl,bl);
    stree.prepare_sph();
    MutualInteract<neighbour_finder,sticky_cell,sticky_soul>
      (stree.root(),stree.depth(),&sfind);
    na = sfind.actual_size_of_list();
  } else {                     // sticky particles
    LoopSouls(tree_type,TREE,Si) if(is_sticky(FL[mybody(Si)])) {
      flag_for_subtree(Si);
      Si->add_body_flag(FL);
    }
    sticky_tree   stree(TREE,s,v);
    sticky_finder sfind(tau,nl,bl);
    stree.prepare_sticky();
    MutualInteract<sticky_finder,sticky_cell,sticky_soul>
      (stree.root(),stree.depth(),&sfind);
    na = sfind.actual_size_of_list();
  }
}
//-----------------------------------------------------------------------------+
void falcON::count_sph_neighbours(const areal* s, areal* NUM)
{
  if(BODIES) error(" mismatch in using bodies/arrays");
  LoopSouls(tree_type,TREE,Si) if(is_sph(FL[mybody(Si)])) {
    flag_for_subtree(Si);
    Si->add_body_flag(FL);
  }
  sticky_tree       stree(TREE,s,0);
  neighbour_counter scount;
  stree.prepare_sph();
  MutualInteract<neighbour_counter,sticky_cell,sticky_soul>
    (stree.root(),stree.depth(),&scount);
  stree.set_num(NUM);
}
////////////////////////////////////////////////////////////////////////////////
